import random
import textwrap

import networkx as nx
import numpy as np
import pytest

import pelita.maze_generator as mg

SEED = 103525239


@pytest.fixture()
def set_seed():
    random.seed(SEED)

def test_maze_bytes_str_conversions():
    # note that the first empty line is needed!
    maze_str = """
                  ##################
                  #. ... .##.     y#
                  # # #  .  .### #x#
                  # # ##.   .      #
                  #      .   .## # #
                  #a# ###.  .  # # #
                  #b     .##. ... .#
                  ##################"""
    maze_bytes = bytes(maze_str, 'ascii')
    maze_clist = [[b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#'],
                  [b'#',b'.',b' ',b'.',b'.',b'.',b' ',b'.',b'#',b'#',b'.',b' ',b' ',b' ',b' ',b' ',b'y',b'#'],
                  [b'#',b' ',b'#',b' ',b'#',b' ',b' ',b'.',b' ',b' ',b'.',b'#',b'#',b'#',b' ',b'#',b'x',b'#'],
                  [b'#',b' ',b'#',b' ',b'#',b'#',b'.',b' ',b' ',b' ',b'.',b' ',b' ',b' ',b' ',b' ',b' ',b'#'],
                  [b'#',b' ',b' ',b' ',b' ',b' ',b' ',b'.',b' ',b' ',b' ',b'.',b'#',b'#',b' ',b'#',b' ',b'#'],
                  [b'#',b'a',b'#',b' ',b'#',b'#',b'#',b'.',b' ',b' ',b'.',b' ',b' ',b'#',b' ',b'#',b' ',b'#'],
                  [b'#',b'b',b' ',b' ',b' ',b' ',b' ',b'.',b'#',b'#',b'.',b' ',b'.',b'.',b'.',b' ',b'.',b'#'],
                  [b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#',b'#']]
    maze_arr = np.array(maze_clist, dtype=bytes)
    assert np.all(mg.bytes_to_maze(maze_bytes) == maze_arr)
    assert np.all(mg.str_to_maze(maze_str) == maze_arr)
    # round trip
    # we must dedent the string and remove the first newline
    # we have the newline in the first line to have dedent work out of the box
    maze_str = textwrap.dedent(maze_str[1:])
    maze_bytes = bytes(maze_str, 'ascii')
    assert mg.maze_to_bytes(maze_arr) == maze_bytes
    assert mg.maze_to_str(maze_arr) == maze_str


def test_create_half_maze(set_seed):
    # this test is not really testing that create_half_maze does a good job
    # we only test that we keep in returning the same maze when the random
    # seed is fixed, in case something changes during future porting/refactoring
    maze_str = """################################
                  #         #    #               #
                  #  # #########                 #
                  #  # #                         #
                  #  #         # #               #
                  #  ####### ###                 #
                  #  #                           #
                  # ### ###### ###               #
                  #      #       #               #
                  ##### ## # ###                 #
                  #   #        #                 #
                  #   #          #               #
                  ##    ###### #                 #
                  #   #      #                   #
                  #              #               #
                  ################################"""

    maze = mg.empty_maze(16,32)
    mg.create_half_maze(maze, 8)
    expected = mg.str_to_maze(maze_str)
    assert np.all(maze == expected)

def test_conversion_to_nx_graph():
    maze_str = """##################
                  #       ##       #
                  # # #      ### # #
                  # # ##           #
                  #           ## # #
                  # # ###      # # #
                  #       ##       #
                  ##################"""
    maze = mg.str_to_maze(maze_str)
    graph = mg.walls_to_graph(maze)
    # now derive a maze from a graph manually
    # - start with a maze full of walls
    newmaze = mg.empty_maze(maze.shape[0], maze.shape[1])
    newmaze.fill(mg.W)
    # - loop through each node of the graph and remove a wall at the
    # corresponding coordinate
    for node in graph.nodes():
        newmaze[node[1], node[0]] = mg.E
    assert np.all(maze == newmaze)

def test_find_one_dead_end():
    # this maze has exactly one dead end at coordinate (1,1)
    maze_dead = """########
                   # #    #
                   #      #
                   #      #
                   ########"""

    maze = mg.str_to_maze(maze_dead)
    graph = mg.walls_to_graph(maze)
    width = maze.shape[1]
    dead_ends = mg.find_dead_ends(graph, width)
    assert len(dead_ends) == 1
    assert dead_ends[0] == (1,1)

def test_find_multiple_dead_ends(set_seed):
    # this maze has exactly three dead ends at coordinates (1,1), (1,5), (3,5)
    maze_dead = """############
                   # #        #
                   #          #
                   #          #
                   # # #      #
                   # # #      #
                   ############"""

    maze = mg.str_to_maze(maze_dead)
    graph = mg.walls_to_graph(maze)
    width = maze.shape[1]
    dead_ends = mg.find_dead_ends(graph, width)
    assert len(dead_ends) == 3
    dead_ends.sort()
    assert dead_ends[0] == (1,1)
    assert dead_ends[1] == (1,5)
    assert dead_ends[2] == (3,5)

def test_find_multiple_dead_ends_on_the_right(set_seed):
    # this maze has exactly three dead ends at coordinates (10,1), (10,5), (8,5)
    maze_dead = """############
                   #        # #
                   #          #
                   #          #
                   #      # # #
                   #      # # #
                   ############"""

    maze = mg.str_to_maze(maze_dead)
    graph = mg.walls_to_graph(maze)
    width = maze.shape[1]
    dead_ends = mg.find_dead_ends(graph, width)
    assert len(dead_ends) == 3
    dead_ends.sort()
    assert dead_ends[2] == (10,5)
    assert dead_ends[1] == (10,1)
    assert dead_ends[0] == (8,5)

def test_remove_one_dead_end():
    # this maze has exactly one dead end at coordinate (1,1)
    maze_dead = """########
                   # #    #
                   #      #
                   #      #
                   ########"""

    maze = mg.str_to_maze(maze_dead)
    mg.remove_dead_end((1,1), maze)
    assert maze[1,1] == mg.E

def test_remove_multiple_dead_ends(set_seed):
    # this maze has exactly three dead ends at coordinates (1,1), (1,5), (3,5)
    maze_dead = """############
                   # #        #
                   #          #
                   #          #
                   # # #      #
                   # # #      #
                   ############"""

    maze = mg.str_to_maze(maze_dead)
    mg.remove_all_dead_ends(maze)
    # There are many ways of getting rid of the two dead ends at the bottom
    # The one that requires the less work is to remove the left-bottom wall
    # This one is the solution we get from remove_all_dead_ends, but just
    # because the order in which we find dead ends is from top to bottom and from
    # left to right.
    # In other words, remove the dead ends here actually means getting this maze back
    expected_maze = """############
                       #          #
                       #          #
                       #          #
                       # # #      #
                       #   #      #
                       ############"""
    expected_maze = mg.str_to_maze(expected_maze)
    assert np.all(maze == expected_maze)

def test_find_chamber():
    # This maze has one single chamber, whose entrance is one of the
    # nodes (1,2), (1,3) or (1,4)
    maze_chamber = """############
                      #   #      #
                      #   #      #
                      # ###      #
                      #          #
                      #          #
                      ############"""

    maze = mg.str_to_maze(maze_chamber)
    maze_orig = maze.copy()
    mg.remove_all_dead_ends(maze)
    # first, check that the chamber is not mistaken for a dead end
    assert np.all(maze_orig == maze)
    # now check that we detect it
    graph = mg.walls_to_graph(maze)
    # there are actually two nodes that can be considered entrances
    entrance, chamber = mg.find_chamber(graph)
    assert entrance in ((1,2), (1,3), (1,4))
    # check that the chamber contains the right nodes. Convert to set, because
    # the order is irrelevant
    if entrance == (1,4):
        expected_chamber = {(1,1), (1,2), (1,3), (2,1), (2,2), (3,1), (3,2)}
    elif entrance == (1,3):
        expected_chamber = {(1,1), (1,2), (2,1), (2,2), (3,1), (3,2)}
    else:
        expected_chamber = {(1,1), (2,1), (2,2), (3,1), (3,2)}
    assert set(chamber) == expected_chamber

    # now remove the chamber and verify that we don't detect anything
    # we just remove wall (4,1) manually
    maze = mg.str_to_maze(maze_chamber)
    maze[1,4] = mg.E # REMEMBER! Indexing is maze[y,x]!!!
    graph = mg.walls_to_graph(maze)
    entrance, chamber = mg.find_chamber(graph)
    assert entrance is None
    assert chamber == []




maze_one_chamber = """############
                      #   #      #
                      #   #      #
                      # ###      #
                      #          #
                      #          #
                      ############"""
maze_two_chambers = """############
                       #   #      #
                       #   #      #
                       # ###  # ###
                       #      #   #
                       #      #   #
                       #      #   #
                       #      #   #
                       ############"""
maze_neighbor_chambers = """####################
                            #   ##   #         #
                            #   ##   #         #
                            # ###### #         #
                            #                  #
                            #                  #
                            #                  #
                            ####################"""
maze_chamber_in_chamber = """######################
                             #                    #
                             #                    #
                             #                    #
                             ##### ####           #
                             #        #           #
                             #   #    #           #
                             ######################"""
maze_chamber_bonanza = """################################
                          #   #   #                      #
                          #   #   #                      #
                          #   ### #                      #
                          #                              #
                          #                              #
                          ###### ##########              #
                          #               #              #
                          #               #              #
                          # ###########   #              #
                          #           #                  #
                          ##### ####  #   #              #
                          #        #  #   #              #
                          #   #    #  #   #              #
                          ################################"""


@pytest.mark.parametrize("maze_chamber", (maze_one_chamber,
                                          maze_two_chambers,
                                          maze_neighbor_chambers,
                                          maze_chamber_in_chamber,
                                          maze_chamber_bonanza,))
def test_remove_all_chambers(set_seed, maze_chamber):
    maze = mg.str_to_maze(maze_chamber)
    mg.remove_all_chambers(maze)
    # there are no more chambers if the connectivity of the graph is larger than 1
    graph = mg.walls_to_graph(maze)
    assert nx.node_connectivity(graph) > 1
    #XXX: TODO -> an easy way of getting rid of chambers is to just remove all the
    # walls. How can we test that this is not what we are doing but that instead
    # we are removing just a few walls?

@pytest.mark.parametrize('iteration', range(1,11))
def test_get_new_maze(set_seed, iteration):
    # generate a few mazes and check them for consistency
    local_seed = random.randint(1,2**31-1)*iteration
    maze_str = mg.get_new_maze(8,16,nfood=15,seed=local_seed)
    maze = mg.str_to_maze(maze_str)
    height, width = maze.shape
    # check that the returned maze has all the pacmen
    for pacman in (b'a',b'x',b'b',b'y'):
        assert np.any(maze == pacman)
        # now that we now we have a pacman, check that we have it only once
        # and remove it by putting an empty space instead
        row, col = np.nonzero(maze == pacman)
        assert len(row) == 1
        assert len(col) == 1
        maze[row,col] = mg.E

    # check that we have in total twice nfood in the maze
    assert (maze == mg.F).sum() == 15*2
    # remove the food for computing dead ends and chambers
    maze[maze == mg.F] = mg.E

    # check that the returned maze is center-mirror symmetric
    left_maze = np.flipud(np.fliplr(maze[:,width//2:]))
    assert np.all(left_maze == maze[:,:width//2])

    # check that we don't have any dead ends
    # no node in the graph should have only one connection
    graph = mg.walls_to_graph(maze)
    for node, degree in graph.degree():
        assert degree > 1

    # now check that we don't have chambers, i.e. the connectivity of the
    # graph is > 1
    assert nx.node_connectivity(graph) > 1

def test_odd_width():
    with pytest.raises(ValueError):
        mg.get_new_maze(2,3,1)

def test_add_food():
    mini = """########
              #      #
              #      #
              ########"""
    maze = mg.str_to_maze(mini)
    # We have 12 empty squares in total, so 6 on the left side.
    # -> 2 slots are taken by the dividing border, so we have 4 left
    # -> 2 are reserved for pacmen, so we have only 2 slots for food
    # - check that we can indeed accommodate 2 pellets in the maze
    lmaze = maze.copy()
    mg.add_food(lmaze, 2)
    assert (lmaze == mg.F).sum() == 2
    # - check that if we add a wall in a free spot that is not reserved by the
    #   pacmen we can only accommodate 1 pellet
    lmaze = maze.copy()
    lmaze[1,2] = mg.W
    mg.add_food(lmaze, 1)
    assert (lmaze == mg.F).sum() == 1
    # - if we try to add more food, we complain
    lmaze = maze.copy()
    lmaze[1,2] = mg.W
    with pytest.raises(ValueError):
        mg.add_food(lmaze, 2)
    # - check that if no space is left we complain
    lmaze = maze.copy()
    lmaze[1,2], lmaze[2,2] = mg.W, mg.W
    with pytest.raises(ValueError):
        mg.add_food(lmaze, 1)
    # - check that we fail if we get passed unreasonable amounts of food
    lmaze = maze.copy()
    with pytest.raises(ValueError):
        mg.add_food(lmaze, -1)
    # - check that we can cope with no food to add
    lmaze = maze.copy()
    mg.add_food(lmaze, 0)
    assert np.all(lmaze == maze)




#!/usr/bin/env python3

import argparse
import cProfile
import functools
import subprocess
import timeit

from pelita.layout import parse_layout
from pelita.game import run_game
from pelita.player import stopping_player, nq_random_player

LAYOUT="""
##################################
#...   #      .#     #  #       y#
# ## #   # ###    #  #     #####x#
#.   # #    # .   # ##           #
#.#    #  .    #    .  # #########
# ## # ## ####    # ##.   . .   .#
#.. .  .  #. . #. #  # ## #####  #
# ## #### #.## #     #  .  . . ..#
#..  ..   # #  #  #    ##### #####
##### #####    #  #  # #   ..  ..#
#.. . .  .  #     # ##.# #### ## #
#  ##### ## #  # .# . .#  .  . ..#
#.   . .   .## #    #### ## # ## #
######### #  .    #    .  #    #.#
#           ## #   . #    # #   .#
#a#####     #  #    ### #   # ## #
#b       #  #     #.      #   ...#
##################################
"""

layout = parse_layout(LAYOUT)

def run(teams, max_rounds):
    return run_game(teams, max_rounds=max_rounds, layout_dict=layout, print_result=False, allow_exceptions=True, store_output=subprocess.DEVNULL)

def parse_args():
    parser = argparse.ArgumentParser(description='Benchmark pelita run_game')
    parser.add_argument('--repeat', help="Number of repeats of timeit.", default=5, type=int)
    parser.add_argument('--number', help="Number of iterations inside timeit.", default=10, type=int)
    parser.add_argument('--max-rounds', help="Max rounds.", default=300, type=int)

    parser.add_argument('--cprofile', help="Show cProfile output with test teams (int).", default=None, type=int)

    return parser.parse_args()

if __name__ == '__main__':
    args = parse_args()
    REPEAT = args.repeat
    NUMBER = args.number
    MAX_ROUNDS = args.max_rounds

    tests = [
        ("Stopping", [stopping_player, stopping_player]),
        ("NQ_Random", [nq_random_player, nq_random_player]),
        ("Stopping (remote)", ["pelita/player/StoppingPlayer.py", "pelita/player/StoppingPlayer.py"]),
        ("NQ_Random (remote)", ["pelita/player/RandomPlayers.py", "pelita/player/RandomPlayers.py"]),
    ]

    if args.cprofile is None:
        print(f"Running {NUMBER} times with max {MAX_ROUNDS} rounds. Fastest out of {REPEAT}:")

        for name, teams in tests:
            result = min(timeit.repeat(functools.partial(run, teams=teams, max_rounds=MAX_ROUNDS), repeat=REPEAT, number=NUMBER))
            print(f"{name:<20}: {result}")

    else:
        name, teams = tests[args.cprofile]
        print(f"Running cProfile for teams {name} with max {MAX_ROUNDS} rounds:")
        max_rounds = MAX_ROUNDS
        cProfile.runctx("""run(teams, max_rounds)""", globals(), locals())




def random_explorer_player(bot, state):
    """ Least visited random player. Will prefer moving to a position it’s never seen before. """

    if not bot.turn in state:
        # initialize bot
        state[bot.turn] = { 'visited': [] }

    if bot.position in state[bot.turn]['visited']:
        state[bot.turn]['visited'].remove(bot.position)
    state[bot.turn]['visited'].insert(0, bot.position)

    # possible candidates
    positions = bot.legal_positions[:]
    # go through all visited positions and remove them
    # from our candidate list
    for pos in state[bot.turn]['visited']:
        if len(positions) == 1:
            # only one position left, we’ll take it
            return positions[0]
        if len(positions) == 0:
            return bot.position
        if pos in positions:
            positions.remove(pos)

    # more than one move left
    return bot.random.choice(positions)


TEAM_NAME = "Random Explorer Players"
move = random_explorer_player



import pytest

import itertools
from pathlib import Path
from textwrap import dedent

from pelita.layout import *


def test_get_available_layouts():
    available = get_available_layouts(size='all')
    assert 1200 == len(available)
    # now also test the filter
    available = get_available_layouts(size='normal')
    assert 1000 == len(available)

def test_get_layout_by_name():
    target_layout = """
################
# .. #   .  # y#
# ..   ##.. # x#
#  ###   ..  . #
# .  ..   ###  #
#a # ..##   .. #
#b #  .   # .. #
################
"""
    loaded = get_layout_by_name('small_000')
    assert target_layout.strip() == loaded.strip()

def test_get_random_layout():
    fails = 0
    for i in range(10):
        l1 = get_random_layout()
        l2 = get_random_layout()
        if l1 == l2:
            fails += 1
    assert fails < 10, "Can't get random layouts!"

def test_get_random_layout_returns_correct_layout():
    name, layout = get_random_layout()
    layout2 = get_layout_by_name(name)
    assert layout == layout2

def test_get_random_layout_random_seed():
    name, layout = get_random_layout(size='small', seed=1)
    assert name == 'small_017'

def test_legal_layout():
    layout = """
             ######
             # . y#
             #. #x#
             #a# .#
             #b . #
             ######
             """
    parsed_layout = parse_layout(layout)
    ewalls = set()
    for x in range(6):
        for y in range(6):
            if (x == 0 or x == 5) or (y == 0 or y == 5):
                ewalls.add((x,y))
    ewalls.update([(3, 2),(2, 3)])
    efood = sorted([(2, 1), (1, 2), (4, 3), (3, 4)])
    ebots = [(1, 3), (4, 2), (1, 4), (4, 1)]
    assert parsed_layout['walls'] == tuple(sorted(ewalls))
    assert parsed_layout['food'] == efood
    assert parsed_layout['bots'] == ebots
    assert parsed_layout['shape'] == (6, 6)
    assert wall_dimensions(parsed_layout['walls']) == parsed_layout['shape']

def test_legal_layout_with_added_items():
    layout = """
             ######
             # . y#
             #. #x#
             #a# .#
             #  . #
             ######
             """
    added_food = [(1,1), (4,4)]
    added_bots = {'b': (1,4)}
    parsed_layout = parse_layout(layout, food=added_food, bots=added_bots)
    ewalls = set()
    for x in range(6):
        for y in range(6):
            if (x == 0 or x == 5) or (y == 0 or y == 5):
                ewalls.add((x,y))
    ewalls.update([(3, 2),(2, 3)])
    efood = sorted([(2, 1), (1, 2), (4, 3), (3, 4)]+added_food)
    ebots = [(1, 3), (4, 2), (1, 4), (4, 1)]
    assert parsed_layout['walls'] == tuple(sorted(ewalls))
    assert parsed_layout['food'] == efood
    assert parsed_layout['bots'] == ebots
    assert parsed_layout['shape'] == (6, 6)
    assert wall_dimensions(parsed_layout['walls']) == parsed_layout['shape']

def test_hole_in_horizontal_border():
    layout = """
             ###### #########
             # ..       .. y#
             #. ######..  #x#
             #  .  .   .  # #
             # #  .   .  .  #
             #a#  ..###### .#
             #b ..       .. #
             ################
             """
    with pytest.raises(ValueError, match=r"Layout must be enclosed by walls.*"):
        parse_layout(layout)

def test_odd_width():
    layout = """
             ###############
             #  .      .. y#
             #. #####..  #x#
             #  .     .  # #
             # #  .  .  .  #
             #a#  .###### .#
             #b ..      .. #
             ###############
             """
    with pytest.raises(ValueError, match=r"Layout width must be even.*"):
        parse_layout(layout)

def test_different_widths():
    layout = """
             ################
             # ..        .. y#
             #. ######..  #x#
             #  .  .   .  # #
             # #  .   .  .  #
             #a#  ..###### .#
             #b ..       .. #
             ################
             """
    with pytest.raises(ValueError, match=r"Layout rows have differing widths.*"):
        parse_layout(layout)

def test_hole_in_vertical_border():
    layout = """
             ################
             # ..       .. y#
             #. ######..  # #
             #  .  .   .  # x
             # #  .   .  .  #
             #a#  ..###### .#
             #b ..       .. #
             ################
             """
    with pytest.raises(ValueError, match=r"Layout must be enclosed by walls.*"):
        parse_layout(layout)

def test_last_row_not_complete():
    layout = """
             ################
             # ..       .. y#
             #. ######..  #x#
             #  .  .   .  # #
             # #  .   .  .  #
             #a#  ..###### .#
             #b ..       .. #
             """
    with pytest.raises(ValueError, match=r"Layout must be enclosed by walls.*"):
        parse_layout(layout)

def test_twice_the_same_bot():
    layout = """
             ################
             # ..       .. y#
             #. ######..  #y#
             #  .  .   .  # #
             # #  .   .  .  #
             #a#  ..###### .#
             #b ..       .. #
             ################
             """
    with pytest.raises(ValueError, match=r"Cannot set bot y to \(14, 2\) .*"):
        parse_layout(layout)

def test_missing_one_bot():
    layout = """
             ################
             # ..       .. y#
             #. ######..  #x#
             #  .  .   .  # #
             # #  .   .  .  #
             #a#  ..###### .#
             #  ..       .. #
             ################
             """
    with pytest.raises(ValueError, match=r".* ['b']"):
        parse_layout(layout)

def test_broken_added_food():
    layout = """
             ######
             # . y#
             #. #x#
             #a# .#
             #b . #
             ######
             """
    added_food = [(10,10)]
    with pytest.raises(ValueError, match=r"food item at \(10, 10\) is .*"):
        parsed_layout = parse_layout(layout, food=added_food)
    added_food = [(2,3)]
    with pytest.raises(ValueError, match=r"food item at \(2, 3\) is .*"):
        parsed_layout = parse_layout(layout, food=added_food)

def test_broken_added_bot():
    layout = """
             ######
             # . y#
             #. #x#
             # # .#
             #b . #
             ######
             """
    added_bots = {'a': (10,10)}
    with pytest.raises(ValueError, match=r"bot a at \(10, 10\) is .*"):
        parsed_layout = parse_layout(layout, bots=added_bots)
    added_bots = {'a':(2,3)}
    with pytest.raises(ValueError, match=r"bot a at \(2, 3\) is .*"):
        parsed_layout = parse_layout(layout, bots=added_bots)

def test_override_bot():
    layout = """
             ######
             # . y#
             #. #x#
             #a# .#
             #b . #
             ######
             """
    added_bots = {'a': (1,1)}
    parsed_layout = parse_layout(layout, bots=added_bots)
    assert parsed_layout['bots'][0] == (1,1)

def test_wrong_bot_names():
    layout = """
             ######
             # . y#
             #. #x#
             #a# .#
             #b . #
             ######
             """
    added_bots = {'e': (1,1)}
    with pytest.raises(ValueError, match=r"Invalid Bot names in .*"):
        parsed_layout = parse_layout(layout, bots=added_bots)

def test_wrong_bot_names_2():
    layout = """
             ######
             # . e#
             #. #x#
             #a# .#
             #b . #
             ######
             """
    added_bots = {'y': (1,1)}
    with pytest.raises(ValueError, match=r"Unknown character.*"):
        parsed_layout = parse_layout(layout, bots=added_bots)

def test_roundtrip():
    input_layout =  """ ########
                        #a  .  #
                        #b    x#
                        #  .  y#
                        ########
                        """

    expected_layout = \
"""########
#a  .  #
#b    x#
#  .  y#
########
"""
    layout = parse_layout(input_layout)
    out = layout_as_str(**layout)
    assert out == dedent(expected_layout)
    layout = parse_layout(out)
    out = layout_as_str(**layout)
    assert out == dedent(expected_layout)


def test_incomplete_roundtrip():
    # We create a layout where a, b and x, y sit on top of each other and on a food pellet.
    input_layout =  """ ########
                        #b  .  #
                        #      #
                        #  .  x#
                        ########
                        """
    ab_bots = (1, 1)
    xy_bots = (6, 3)
    bots = {'a': ab_bots, 'b': ab_bots, 'x': xy_bots, 'y': xy_bots }
    food = [ab_bots, xy_bots]

    # In the layout_as_str, a and x will be shown but nothing else
    expected_layout = \
"""########
#a  .  #
#      #
#  .  x#
########
"""
    layout = parse_layout(input_layout, bots=bots, food=food)
    out = layout_as_str(**layout)
    assert out == dedent(expected_layout)


def test_empty_lines():
    simple_layout_1 = (
        """ ####
            #ax#
            #by#
            #. #
        #### """)

    simple_layout_2 = (
        """

            ####
            #ax#
            #by#
            #. #
            ####

            """)
    layout1 = parse_layout(simple_layout_1)
    layout2 = parse_layout(simple_layout_2)
    assert layout1 == layout2


@pytest.mark.parametrize('pos, legal_positions', [
    ((2, 2), {(2, 1), (2, 3), (1, 2), (3, 2), (2, 2)}),
    ((1, 1), {(1, 2), (2, 1), (1, 1)}),
    ((4, 2), {(4, 2), (4, 1), (4, 3), (3, 2)}),
    ((4, 1), {(4, 2), (4, 1)})
])
def test_legal_positions(pos, legal_positions):
    test_layout = (
        """ ######
            #a # #
            #b   #
            #xy  #
            ###### """)
    parsed = parse_layout(test_layout)
    assert set(get_legal_positions(parsed['walls'], parsed['shape'], pos)) == legal_positions


@pytest.mark.parametrize('pos', [
    (0, 0),
    (0, 1),
    (-1, 0),
    (7, 7),
    (3, 1)
])
def test_legal_positions_fail(pos):
    test_layout = (
        """ ######
            #a # #
            #b   #
            #yx  #
            ###### """)
    parsed = parse_layout(test_layout)
    with pytest.raises(ValueError):
        get_legal_positions(parsed['walls'], parsed['shape'], pos)


def test_load():
    layout1="""
        ########
        # ###ya#
        #bx ...#
        ########
    """
    layout = parse_layout(layout1)
    assert layout['bots'] == [(6, 1), (2, 2), (1, 2),(5, 1)]

def test_bots_in_same_position():
    layout_str = """
        ########
        # ###  #
        # . ...#
        ########
    """
    bot_dict = {"a": (1, 1),
                "b": (1, 1),
                "x": (1, 1),
                "y": (1, 1)}
    layout = parse_layout(layout_str, bots=bot_dict)
    assert layout['bots'] == [(1, 1), (1, 1), (1, 1),(1, 1)]


# All combinations of bots that can be switched on/off
@pytest.mark.parametrize('bots_hidden', itertools.product(*[(True, False)] * 4))
def test_parse_layout_game_bad_number_of_bots(bots_hidden):
    """ parse_layout should fail when a wrong number of bots is given. """
    test_layout = """
        ##################
        #a#.  .  # .     #
        #b#####    #####y#
        #     . #  .  .#x#
        ################## """
    # remove bot i when bots_hidden[i] is True
    for char, idx in BOT_N2I.items():
        if bots_hidden[idx]:
            test_layout = test_layout.replace(char, ' ')

    if list(bots_hidden) == [False] * 4:
        # no bots are hidden. it should succeed
        parsed_layout = parse_layout(test_layout)
    else:
        with pytest.raises(ValueError):
            parsed_layout = parse_layout(test_layout)



""" Signal handlers.
"""

import logging
import os
import platform
import signal

_logger = logging.getLogger(__name__)

def exit_handler(*args):
    """ Simple handler to just quit using os._exit(-1).

    This is not a very graceful approach. None of the resources that the program
    acquired during its lifecycle will be released. This may or may not affect
    network resources, open files, allocated memory, and possibly even child
    processes.

    """
    # defensively shutdown the logging system
    logging.shutdown()
    # kill all processes in processgroup.
    # we need this in case we use the multiprocessing.
    if platform.system() == 'Windows':
        import win32api, win32con
        handle = win32api.OpenProcess(win32con.PROCESS_TERMINATE,
                                      0, os.getpid())
        win32api.TerminateProcess( handle, 0 )
        win32api.CloseHandle( handle )
    else:
        os.kill(os.getpid(), signal.SIGTERM)


# is added  pelita/ui/tk_canvas:TkApplication
def wm_delete_window_handler():
    _logger.info("Got WM_DELETE_WINDOW. Exit!")
    exit_handler()



"""
Generate maze layouts for 'pelita', without dead ends.

Algorithm:
* start with an empty grid
* draw a wall with gaps, dividing the grid in 2
* repeat recursively for each sub-grid
* find dead ends
* remove a wall at the dead ends

Players 1,3 always start in the bottom left; 2,4 in the top right
Food is placed randomly (though not too close to the pacmen starting positions)

Notes:
the final map includes a symmetric, flipped copy
the first wall has k gaps, the next wall has k/2 gaps, etc. (min=1)

Inspired by code by Dan Gillick
Completely rewritten by Pietro Berkes
Rewritten again (but not completely) by Tiziano Zito
"""
import random

import numpy as np
import networkx as nx


north = (0, -1)
south = (0, 1)
east = (1, 0)
west = (-1, 0)


# character constants for walls, food, and empty spaces
W = b'#'
F = b'.'
E = b' '


def empty_maze(height, width):
    """Return an empty maze with external walls.

    A maze is a 2D array of characters representing walls, food, and agents.
    An empty maze is made of empty tiles, except for the external walls.
    """

    maze = np.empty((height, width), dtype='c')
    maze.fill(E)

    # add external walls
    maze[0, :].fill(W)
    maze[-1, :].fill(W)
    maze[:, 0].fill(W)
    maze[:, -1].fill(W)

    return maze


def maze_to_bytes(maze):
    """Return bytes representation of maze."""
    lines = [b''.join(maze[i,:])
             for i in range(maze.shape[0])]
    return b'\n'.join(lines)

def maze_to_str(maze):
    """Return a ascii-string representation of maze."""
    bytes_ = maze_to_bytes(maze)
    return bytes_.decode('ascii')

def bytes_to_maze(bytes_):
    """Return a maze numpy bytes array from a bytes representation."""
    rows = []
    for line in bytes_.splitlines():
        line = line.strip()
        if len(line) == 0:
            # skip empty lines
            continue
        cols = []
        for idx in range(len(line.strip())):
            # this crazyness is needed because bytes do not iterate like
            # strings: see the comments about iterating over bytes in
            # https://docs.python.org/3/library/stdtypes.html#bytes-objects
            cols.append(line[idx:idx+1])
        rows.append(cols)

    maze = np.array(rows, dtype=bytes)
    return maze

def str_to_maze(str_):
    """Return a maze numpy bytes array from a ascii string representation."""
    bytes_maze = str_.encode('ascii')
    return bytes_to_maze(bytes_maze)

def create_half_maze(maze, ngaps_center):
    """Fill the left half of the maze with random walls.

    The second half can be created by mirroring the left part using
    the 'complete_maze' function.
    """

    # first, we need a wall in the middle

    # the gaps in the central wall have to be chosen such that they can
    # be mirrored
    ch = maze.shape[0] - 2
    candidates = list(range(ch//2))
    random.shuffle(candidates)
    half_gaps_pos = candidates[:ngaps_center // 2]
    gaps_pos = []
    for pos in half_gaps_pos:
        gaps_pos.append(pos)
        gaps_pos.append(ch - pos - 1)

    # make wall
    _add_wall_at(maze, (maze.shape[1] - 2) // 2 - 1, ngaps_center,
                 vertical=True, gaps_pos=gaps_pos)

    # then, fill the left half with walls
    _add_wall(maze[:, :maze.shape[1] // 2], ngaps_center // 2, vertical=False)

def _add_wall_at(maze, pos, ngaps, vertical, gaps_pos=None):
    """
    add a wall with gaps

    maze -- maze where to place wall, plus a border of one element
    pos -- position where to put the wall within the center of the maze
           (border excluded)
    """

    if not vertical:
        maze = maze.T

    center = maze[1:-1, 1:-1]
    ch, cw = center.shape

    # place wall
    center[:, pos].fill(W)

    # place gaps
    ngaps = max(1, ngaps)
    # choose position of gaps if necessary
    if gaps_pos is None:
        # choose aandom positions
        gaps_pos = list(range(ch))
        random.shuffle(gaps_pos)
        gaps_pos = gaps_pos[:ngaps]
        # do not block entrances
        if maze[0][pos + 1] == E:
            gaps_pos.insert(0, 0)
        if maze[-1][pos + 1] == E:
            gaps_pos.insert(0, ch - 1)
    for gp in gaps_pos:
        center[gp, pos] = E

    sub_mazes = [maze[:, :pos + 2], maze[:, pos + 1:]]

    if not vertical:
        sub_mazes = [sm.T for sm in sub_mazes]

    return sub_mazes

def _add_wall(maze, ngaps, vertical):
    """Recursively build the walls of the maze.

    grid -- 2D array of characters representing the maze
    ngaps -- number of empty spaces to leave in the wall
    vertical -- if True, create a vertical wall, otherwise horizontal
    """

    h, w = maze.shape
    center = maze[1:-1, 1:-1]
    ch, cw = center.shape

    # no space for walls, interrupt recursion
    if ch < 3 and cw < 3:
        return

    size = cw if vertical else ch
    # create a wall only if there is some space in this direction
    min_size = random.randint(3, 5)
    if size >= min_size:
        # place the wall at random spot
        pos = random.randint(1, size-2)
        sub_mazes = _add_wall_at(maze, pos, ngaps, vertical)

        # recursively add walls
        for sub_maze in sub_mazes:
            _add_wall(sub_maze, max(1, ngaps // 2), not vertical)


def walls_to_graph(maze):
    """Transform a maze in a graph.

    The data on the nodes correspond to their coordinates, data on edges is
    the actions to take to transition to that edge.

    Returns:
    graph -- a Graph
    first_node -- the first node in the Graph
    """

    h, w = maze.shape
    directions = [west, east, north, south]

    graph = nx.Graph()
    # define nodes for maze
    for x in range(w):
        for y in range(h):
            if maze[y, x] != W:
                graph.add_node((x,y))
                # this is a free position, get its neighbors too
                for dx, dy in directions:
                    nbx, nby = (x+dx, y+dy)
                    # do not go out of bounds
                    try:
                        if maze[nby, nbx] == E:
                            graph.add_edge((x, y), (nbx, nby))
                    except IndexError:
                        # this move brought us out of the maze, just ignore it
                        continue
    return graph


def find_dead_ends(graph, width):
    """Find dead ends in a graph."""

    dead_ends = []
    for node in graph.nodes():
        if graph.degree(node) == 1 and node[0] < width-1:
            dead_ends.append(node)

    return dead_ends


def remove_dead_end(dead_node, maze):
    """Remove one dead end in a maze."""

    h, w = maze.shape

    # loop through the neighboring positions and remove the first wall we find
    # as long as it is not on the outer border or in the middle of the maze
    # not in the central wall x==w//2-1
    directions = (north, south, east, west)
    for direction in directions:
        nbx = dead_node[0]+direction[0]
        nby = dead_node[1]+direction[1]
        if nbx not in (0,w-1,w//2-1) and nby not in (0,h-1):
            neighbor = maze[nby, nbx]
            if neighbor == W:
                maze[nby, nbx] = E
                break


def remove_all_dead_ends(maze):
    height, width = maze.shape
    while True:
        maze_graph = walls_to_graph(maze)
        dead_ends = find_dead_ends(maze_graph, width)
        if len(dead_ends) == 0:
            break

        remove_dead_end(dead_ends[0], maze)

def find_chamber(graph):
    """Detect chambers (rooms with a single square entrance).

    Return (entrance, chamber), where `entrance` is the node representing the
    entrance to the chamber (None if no chamber is found), and `chamber` is the
    list of nodes within the chamber (empty list if no nodes are in the chamber).

    The entrance to a chamber is a node that when removed from the graph
    will result in the graph to be split into two disconnected graphs."""
    # minimum_node_cut returns a set of nodes of minimum cardinality that
    # disconnects the graph. This means that we have a chamber if the length
    # of this set is one, i.e. there is one node that when removed disconnects
    # the graph
    cuts = nx.minimum_node_cut(graph)
    if len(cuts) > 1:
        # no chambers, yeah!
        return None, []
    entrance = cuts.pop()
    # remove the cut, i.e. put a wall on the entrance
    lgraph = nx.restricted_view(graph, [entrance],[])
    # now get the resulting subgraphs
    subgraphs = sorted(nx.connected_components(lgraph), key=len)
    # let's get the smallest subgraph: this is going to be a chamber
    # (other subgraphs are other chambers (if any) and the 'rest' of the graph
    # return a list of nodes, instead of a set
    chamber = list(subgraphs[0])
    return entrance, chamber

def get_neighboring_walls(maze, locs):
    """Given a list of coordinates in the maze, return all neighboring walls.

    Walls on the outer border are ignored automatically."""
    height, width = maze.shape
    walls = []
    seen = []
    for nodex, nodey in locs:
        # if we are already on the border, skip this node
        if nodex<=0 or nodex>=(width-1) or nodey<=0 or nodey>=(height-1):
            continue
        # explore all directions around the current node
        for dirx, diry in (north, south, east, west):
            # get coordinates of neighbor in direction (dirx, diry)
            adjx, adjy = nodex+dirx, nodey+diry
            if (adjx, adjy) in seen:
                # we have visited this neighbor already
                continue
            else:
                seen.append((adjx, adjy))
            # check that we still are inside the maze
            if adjx<=0 or adjx>=(width-1) or adjy<=0 or adjy>=(height-1):
                # the neighbor is out of the maze
                continue
            if maze[adjy,adjx] == W:
                # this is a wall, store it
                walls.append((adjx, adjy))
    return walls

def remove_all_chambers(maze):
    maze_graph = walls_to_graph(maze)
    # this will find one of the chambers, if there is any
    entrance, chamber = find_chamber(maze_graph)
    while entrance is not None:
        # get all the walls around the chamber
        walls = get_neighboring_walls(maze, chamber)
        # choose a wall at random among the neighboring one and get rid of it
        bad_wall = random.choice(walls)
        maze[bad_wall[1], bad_wall[0]] = E
        # we may have opened a door into this chamber, but there may be more
        # chambers to get rid of. Or, the wall we picked wasn't good enough and
        # didn't really open a new door to the chamber. I have no idea how to
        # distinguish this two cases. If we knew how to, we would spare quite
        # a few iterations here?
        # Well, as long as we keep on doing this we will eventually get rid
        # of all the chambers
        maze_graph = walls_to_graph(maze)
        entrance, chamber = find_chamber(maze_graph)


def add_food(maze, max_food):
    """Add max_food pellets on the left side of the maze.

    We exclude the pacmen's starting positions and the central dividing border
    """
    if max_food == 0:
        # no food needs to be added, return here
        return
    h, w = maze.shape
    pacmen = [(1,h-2), (1,h-3)]
    # get all free slots on the left side, excluding the dividing border
    free_y, free_x = np.where(maze[:,:w//2-1] == E)
    # convert it to a list of coordinate tuples
    free = list(zip(free_x, free_y))
    # remove the pacmen starting coordinates (we have to check that they are
    # indeed free before try to remove them
    [free.remove(pacman) for pacman in pacmen if pacman in free]
    # check if we have any free slots left
    if len(free) == 0 and max_food > 0:
        raise ValueError(f'No space left for food in maze')
    elif max_food > len(free):
        # check if we can indeed fit so much food in the maze
        raise ValueError(f'Can not fit {max_food} pellet in {len(free)} free slots')
    elif max_food < 0:
        raise ValueError(f'Can not add negative number of food ({max_food} given)')

    # now take max_food random positions out of this list
    food = random.sample(free, max_food)
    # fit it in the maze
    for col, row in food:
        maze[row, col] = F

def add_pacmen(maze):
    ## starting pacmen positions
    maze[-2, 1] = b'b'
    maze[-3, 1] = b'a'
    maze[1, -2] = b'y'
    maze[2, -2] = b'x'

def get_new_maze(height, width, nfood, seed=None):
    """Create a new maze in text format.

    The maze is created with a recursive creation algorithm. The maze part of
    the blue team is a center-mirror version of the one for the red team.

    The function reserves space for 2 PacMan for each team in upper-right
    and lower-left corners of the maze. Food is added at random.

    Input arguments:
    height, width -- the size of the maze, including the outer walls
    nfood -- number of food dots for each team
    seed -- if not None, the random seed used to generate the maze
    """
    if width%2 != 0:
        raise ValueError(f'Width must be even ({width} given)')

    if seed is None:
        seed = random.randint(1, 2 ** 31 - 1)
    random.seed(seed)

    maze = empty_maze(height, width)
    create_half_maze(maze, height // 2)

    # make space for pacman (2 pacman each)
    maze[-2, 1] = E
    maze[-3, 1] = E

    # remove dead ends
    remove_all_dead_ends(maze)

    # remove chambers
    remove_all_chambers(maze)

    # add food
    add_food(maze, nfood)

    # complete right part of maze with mirror copy
    maze[:, width // 2:] = np.flipud(np.fliplr(maze[:, :width // 2]))

    # add pacman
    add_pacmen(maze)

    return maze_to_str(maze)




class FatalException(Exception): # TODO rename to FatalGameException etc
    pass

class NonFatalException(Exception):
    pass

class PlayerTimeout(NonFatalException):
    pass

class PlayerDisconnected(FatalException):
    # unsure, if PlayerDisconnected should be fatal in the sense of that the team loses
    # it could simply be a network error for both teams
    # and it would be random who will be punished
    pass

class NoFoodWarning(Warning):
    """ Warns when a layout has no food during setup. """
    pass



import contextlib
from dataclasses import dataclass
import io
import json
import logging
import os
import random
import re
import shlex
import signal
import subprocess
import sys
import tempfile
import time

import yaml
import zmq

from ..team import make_team
from . import knockout_mode, roundrobin

_logger = logging.getLogger(__name__)
_mswindows = (sys.platform == "win32")


# Number of points a teams gets for matches in the first round
# Probably not worth it to make it options.
POINTS_DRAW = 1
POINTS_WIN = 2

LOGFILE = None


def check_team(team_spec):
    """ Instantiates a team from a team_spec and returns its name """
    team, _zmq_context = make_team(team_spec)
    return team.team_name


@contextlib.contextmanager
def run_and_terminate_process(args, **kwargs):
    """ This serves as a contextmanager around `subprocess.Popen`, ensuring that
    after the body of the with-clause has finished, the process itself (and the
    process’s children) terminates as well.

    On Unix we try sending a SIGTERM before killing the process group but as
    afterwards we only wait on the first child, this means that the grand children
    do not get the chance to properly terminate.
    In cases where the first child has children that should properly close, the
    first child should catch SIGTERM with a signal handler and wait on its children.

    On Windows we send a CTRL_BREAK_EVENT to the whole process group and
    hope for the best. :)
    """

    _logger.debug(f"Executing: {shlex.join(args)}")

    try:
        if _mswindows:
            p = subprocess.Popen(args, **kwargs, creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)
        else:
            p = subprocess.Popen(args, **kwargs, preexec_fn=os.setsid)
        yield p
        p.poll()
        if p.returncode is not None:
            _logger.debug(f"Subprocess exited with {p.returncode}.")
        else:
            _logger.debug(f"Subprocess has not exited yet.")
    finally:
        if _mswindows:
            _logger.debug("Sending CTRL_BREAK_EVENT to {proc} with pid {pid}.".format(proc=p, pid=p.pid))
            os.kill(p.pid, signal.CTRL_BREAK_EVENT)
        else:
            try:
                pgid = os.getpgid(p.pid)
                _logger.debug("Sending SIGTERM to pgid {pgid}.".format(pgid=pgid))
                os.killpg(pgid, signal.SIGTERM) # send sigterm, or ...
                try:
                    # It would be nicer to wait with os.waitid on the process group
                    # but that does not seem to exist on macOS.
                    p.wait(3)
                except subprocess.TimeoutExpired:
                    _logger.debug("Sending SIGKILL to pgid {pgid}.".format(pgid=pgid))
                    os.killpg(pgid, signal.SIGKILL) # send sigkill, or ...
            except ProcessLookupError:
                # did our process group vanish?
                # we try killing only the child process then
                _logger.debug("Sending SIGTERM to pid {pid}.".format(pid=p.pid))
                p.terminate()
                try:
                    p.wait(3)
                except subprocess.TimeoutExpired:
                    _logger.debug("Sending SIGKILL to pid {pid}.".format(pid=p.pid))
                    p.kill()


def call_pelita(team_specs, *, rounds, size, viewer, seed, team_infos=None, write_replay=False, store_output=False):
    """ Starts a new process with the given command line arguments and waits until finished.

    Returns
    =======
    tuple of (game_state, stdout, stderr)
    """
    team1, team2 = team_specs

    if team_infos is None:
        team_infos = [None, None]

    if seed is not None:
        if isinstance(seed, int):
            # convert to str
            seed = str(seed)
        elif isinstance(seed, str):
            # try that it can be converted to int
            try:
                int(seed)
            except ValueError:
                raise ValueError("seed must be an int, a string that can be converted to int or None.")
        else:
            raise ValueError("seed must be an int, a string that can be converted to int or None.")

    ctx = zmq.Context()
    reply_sock = ctx.socket(zmq.PAIR)

    addr = 'tcp://127.0.0.1'
    reply_port = reply_sock.bind_to_random_port(addr)
    reply_addr = 'tcp://127.0.0.1' + ':' + str(reply_port)

    rounds = ['--rounds', str(rounds)] if rounds else []
    size = ['--size', size] if size else []
    viewer = ['--' + viewer] if viewer else []
    seed = ['--seed', seed] if seed else []
    write_replay = ['--write-replay', write_replay] if write_replay else []
    store_output = ['--store-output', store_output] if store_output else []
    append_blue = ['--append-blue', team_infos[0]] if team_infos[0] else []
    append_red = ['--append-red', team_infos[1]] if team_infos[1] else []

    cmd = [sys.executable, '-m', 'pelita.scripts.pelita_main',
           team1, team2,
           '--reply-to', reply_addr,
           *append_blue,
           *append_red,
           *rounds,
           *size,
           *viewer,
           *seed,
           *write_replay,
           *store_output]

    # We need to run a process in the background in order to await the zmq events
    # stdout and stderr are written to temporary files in order to be more portable
    with tempfile.TemporaryFile(mode='w+t') as stdout_buf, tempfile.TemporaryFile(mode='w+t') as stderr_buf:
        # We use the environment variable PYTHONUNBUFFERED here to retrieve stdout without buffering
        with run_and_terminate_process(cmd, stdout=stdout_buf, stderr=stderr_buf,
                                            universal_newlines=True,
                                            env=dict(os.environ, PYTHONUNBUFFERED='x')) as proc:

            poll = zmq.Poller()
            poll.register(reply_sock, zmq.POLLIN)

            final_game_state = None

            while True:
                evts = dict(poll.poll(1000))

                if not evts and proc.poll() is not None:
                    # no more events and proc has finished.
                    # we break the loop
                    break

                socket_ready = evts.get(reply_sock, False)
                if socket_ready:
                    try:
                        game_state = json.loads(reply_sock.recv_string())
                        finished = game_state.get("gameover", None)
                        whowins = game_state.get("whowins", None)
                        if finished:
                            final_game_state = game_state
                            # The game in the subprocess has finished but the process
                            # may still be running.
                            # Give it a little time to finish writing everything.
                            # Once we exit the `with` statement, the subprocess will
                            # be terminated.
                            try:
                                proc.wait(1)
                            except subprocess.TimeoutExpired:
                                # It didn’t terminate in time by itself.
                                # We exit anyway.
                                pass
                            break
                    except ValueError:  # JSONDecodeError
                        pass
                    except KeyError:
                        pass

        stdout_buf.seek(0)
        stderr_buf.seek(0)
        return (final_game_state, stdout_buf.read(), stderr_buf.read())



def create_team_id(team_id, idx):
    """ Checks that the team_id in the config is valid or else
    creates one from the given index. """
    if team_id is None:
        return "#" + str(idx)
    elif not isinstance(team_id, str):
        raise ValueError("team_id must be string or None.")
    elif not team_id:
        raise ValueError("team_id must not be empty.")
    elif team_id.startswith("#"):
        raise ValueError("team_id must not start with #.")
    else:
        return team_id


@dataclass
class MatchID:
    """Small helper class to keep track of which match is played for better logging."""
    round: int = 1
    match: int = 1
    match_repeat: int = 1
    def __str__(self) -> str:
        if self.match_repeat == 1:
            return f'round{self.round}-match{self.match:02}'
        else:
            return f'round{self.round}-match{self.match:02}-repeat{self.match_repeat}'
    def next_round(self):
        # increase round, reset all other values
        self.round += 1
        self.match = 1
        self.match_repeat = 1
    def next_match(self):
        # increase match, keep round, reset repeat
        self.match += 1
        self.match_repeat = 1
    def next_repeat(self):
        # increase repeat, keep all other values
        self.match_repeat += 1


class Config:
    def __init__(self, config):
        self.teams = {}

        teams = config["teams"]
        # load team names
        for idx, team in enumerate(teams):
            team_id = create_team_id(team.get("id"), idx)
            team_spec = team["spec"]
            team_name = set_name(team_spec)

            if team_id in self.teams:
                raise ValueError("Duplicate team_id {} given.".format(team_id))

            self.teams[team_id] = {
                "spec": team_spec,
                "name": team_name,
                "members": team["members"]
            }

        self.location = config["location"]
        self.date = config["date"]

        self.rounds = config.get("rounds")
        self.size = config.get("size")

        self.viewer = config.get("viewer")
        self.interactive = config.get("interactive")
        self.statefile = config.get("statefile")

        self.greeting = config.get("greeting", "Hello")
        self.farewell = config.get("farewell", "Good evening")
        self.host = config.get("host", "host")

        #: Global random seed.
        #: Keep it fixed or it may be impossible to replicate a tournament.
        #: Individual matches will get a random seed derived from this.
        self.seed = config.get("seed", 42)

        self.bonusmatch = config["bonusmatch"]

        self.speak = config.get("speak")
        self.speaker = config.get("speaker")

        self.tournament_log_folder = None
        self.tournament_log_file = None

    @property
    def team_ids(self):
        return self.teams.keys()

    def team_name(self, team):
        return self.teams[team]["name"]

    def team_group(self, team):
        return f"group {team}"

    def team_name_group(self, team):
        return f"{self.team_name(team)} ({self.team_group(team)})"

    def team_spec(self, team):
        return self.teams[team]["spec"]

    def _print(self, *args, **kwargs):
        print(*args, **kwargs)
        if self.tournament_log_file:
            with open(self.tournament_log_file, 'a') as f:
                kwargs['file'] = f
                print(*args, **kwargs)

    def print(self, *args, **kwargs):
        """Speak while you print. To disable set speak=False.
        You need the program %s to be able to speak.
        Set wait=X to wait X seconds after speaking."""
        if len(args) == 0:
            self._print()
            return
        stream = io.StringIO()
        wait = kwargs.pop('wait', 0.5)
        want_speak = kwargs.pop('speak', None)
        if (want_speak is False) or not self.speak:
            self._print(*args, **kwargs)
        else:
            self._print(*args, file=stream, **kwargs)
            string = stream.getvalue()
            self._print(string, end='')
            sys.stdout.flush()
            self.say(string)
            time.sleep(wait)

    def say(self, string):
        with tempfile.NamedTemporaryFile('w+t') as tmp:
            ansi_seqs = re.compile(r'\x1b[^m]*m')
            string = ansi_seqs.sub('', string)

            tmp.write(string+'\n')
            tmp.flush()
            cmd = shlex.split(self.speaker)
            full_cmd = [*cmd, tmp.name]
            try:
                festival = subprocess.run(full_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True)
            except FileNotFoundError:
                # If we could not find the executable then there is not need to keep on trying.
                # Disabling speak. (Although self.say() will still attempt to speak.)

                print("Could not find executable in call {!r}".format(full_cmd))
                print("Disabling speech synthesis.")

                _logger.warn("Could not find executable in call {!r}".format(full_cmd))
                _logger.warn("Disabling speech synthesis.")

                self.speak = False
                self.wait_for_keypress()
            except subprocess.CalledProcessError as err:
                # A non-zero return value could mean that the call syntax is wrong.
                # However, it could also be a one-time error (maybe the program did
                # not like a character in the string or something like that).
                # We only print the error the first time and only log it afterwards.

                if not getattr(self, '_say_process_error', False):
                    print(err)
                    print("Ignoring this error.")
                    self._say_process_error = True
                    self.wait_for_keypress()
                else:
                    _logger.warn(err)


    def input(self, str, values=None):
        if not values:
            values = []
        while True:
            res = input(str)
            try:
                if res[0] in values:
                    return res[0]
            except IndexError:
                pass


    def wait_for_keypress(self):
        if self.interactive:
            input('--- (press ENTER to continue) ---\n')
        else:
            _logger.debug("Noninteractive. Not asking for keypress.")


class State:
    def __init__(self, config, state=None):
        if state is None:
            self.state = {
                "round1": {
                    "played": [],
                    "unplayed": roundrobin.create_matchplan(config.team_ids)
                },
                "round2": {}
            }
        else:
            self.state = state

    @property
    def round1(self):
        return self.state["round1"]

    @property
    def round2(self):
        return self.state["round2"]

    def save(self, filename):
        if filename:
            with open(filename, 'w') as f:
                yaml.dump(self.state, f, indent=2)

    @classmethod
    def load(cls, config, filename):
        if filename:
            with open(filename) as f:
                return cls(config=config, state=yaml.load(f, Loader=yaml.FullLoader))


def present_teams(config):
    config.wait_for_keypress()
    print("\33[H\33[2J")  # clear the screen

    greeting = config.greeting
    host = config.host

    config.print(f'{greeting} {host}, I am the Python drone. I am here to serve you.', wait=1.5)
    config.print('Welcome to the %s Pelita tournament %s' % (config.location, config.date), wait=1.5)
    config.print('This evening the teams are:', wait=1.5)
    for team_id, team in config.teams.items():
        if isinstance(team_id, int):
            team_id = "#{team_id}".format(team_id=team_id)
        config.print("{team_id}: {team_name}".format(team_id=team_id, team_name=team["name"]))
        for member in team["members"]:
            config.print("{member}".format(member=member), wait=0.1)
        # time.sleep(1)
        config.print()
    config.print('These were the teams. Now you ready for the fight?')


def set_name(team):
    """Get name of team."""
    try:
        return check_team(team)
    except Exception:
        print("*** ERROR: I could not load team {team}. Please help!".format(team=team))
        print(sys.stderr)
        raise


def play_game_with_config(config, teams, *, match_id=None):
    team1, team2 = teams

    if config.tournament_log_folder:
        if match_id:
            log_folder = config.tournament_log_folder / f"{match_id}-{time.strftime('%Y%m%d-%H%M%S')}"
        else:
            log_folder = config.tournament_log_folder / f"match-{time.strftime('%Y%m%d-%H%M%S')}"
        log_folder.mkdir()
        replay_file = log_folder / 'replay'
        log_kwargs = {
            'write_replay': str(replay_file),
            'store_output': str(log_folder)
        }
    else:
        log_folder = None
        log_kwargs = {}

    seed = str(random.randint(0, sys.maxsize))
    team_infos = [config.team_group(team1), config.team_group(team2)]

    res = call_pelita([config.team_spec(team1), config.team_spec(team2)],
                                rounds=config.rounds,
                                size=config.size,
                                viewer=config.viewer,
                                team_infos=team_infos,
                                seed=seed,
                                **log_kwargs)

    if log_folder:
        (_final_state, stdout, stderr) = res
        (log_folder / 'main.out').write_text(stdout)
        (log_folder / 'main.err').write_text(stderr)

    return res


def start_match(config, teams, *, shuffle=False, match_id=None):
    """Start a match between a list of teams. Return the index of the team that won
    False if there was a draw.
    """
    assert len(teams) == 2
    # Should we insist that all teams be different in a match?
    # assert len(teams) == len(set(teams))

    if shuffle:
        random.shuffle(teams)

    team1, team2 = teams

    config.print()
    config.print('Starting match: '+ config.team_name_group(team1)+' vs ' + config.team_name_group(team2))
    config.print()
    config.wait_for_keypress()

    (final_state, stdout, stderr) = play_game_with_config(config, teams, match_id=match_id)
    try:
        whowins = final_state['whowins']

        if whowins == 2:
            config.print('‘{t1}’ and ‘{t2}’ had a draw.'.format(t1=config.team_name(team1),
                                                                t2=config.team_name(team2)))
            return False
        elif whowins == 0 or whowins == 1:
            winner = teams[whowins]
            config.print('‘{team}’ wins'.format(team=config.team_name(winner)))
            return winner
        else:
            raise ValueError
    except (TypeError, ValueError, KeyError):
        config.print("Unable to parse winning result :(")
        config.print("*** ERROR: Apparently the game crashed. At least I could not find the outcome of the game.")
        config.print("*** Maybe stdout helps you to debug the problem")
        config.print(stdout, speak=False)
        config.print("*** Maybe stderr helps you to debug the problem")
        config.print(stderr, speak=False)
        config.print("***", speak=False)
        return None


def start_match_with_replay(config, match, *, shuffle=False, match_id=None):
    """ Runs start_match until it returns a proper output or manual intervention. """

    winner = start_match(config, match, shuffle=shuffle, match_id=match_id)
    while winner is None:
        config.print("Do you want to re-play the game or enter a winner manually?")
        res = config.input("(r)e-play/(0){}/(1){}/(d)raw > ".format(config.team_name(match[0]),
                                                                    config.team_name(match[1])), values="r01d")
        if res == 'r':
            if match_id:
                # increase the repeat count
                match_id.next_repeat()
            winner = start_match(config, match, shuffle=shuffle, match_id=match_id)
        elif res == '0':
            winner = match[0]
        elif res == '1':
            winner = match[1]
        elif res == 'd':
            winner = False
    return winner


def start_deathmatch(config, team1, team2, *, match_id=None):
    """Start a match between team1 and team2 until one of them wins (ie no
    draw.)
    """
    config.print()
    config.print()
    config.print("{} v {}".format(config.team_name(team1), config.team_name(team2)))
    for i in range(3):
        winner = start_match_with_replay(config, [team1, team2], shuffle=True, match_id=match_id)
        config.wait_for_keypress()

        if winner is False or winner is None:
            config.print('Draw -> Now go for a Death Match!')
            if match_id:
                match_id.next_repeat()
            continue
        return winner
    # if we are here, we have no winner after 3 death matches
    # just assign a random winner
    config.print('No winner after 3 Death Matches. Choose a winner at random:', wait=2)
    winner = random.choice((team1, team2))
    config.print('And the winner is', config.team_name(winner))
    return winner


def round1_ranking(config, rr_played):
    import collections
    points = collections.Counter()
    for match in rr_played:
        winner = match["winner"]
        if winner is not False and winner is not None:
            points[match["winner"]] += POINTS_WIN
        else:
            for team in match["match"]:
                points[team] += POINTS_DRAW
    team_points = [(team_id, points[team_id]) for team_id in config.team_ids]
    return sorted(team_points, key=lambda elem: elem[1], reverse=True)


def pp_round1_results(config, rr_played, rr_unplayed, highlight=None):
    if highlight is None:
        highlight = []
    BOLD = '\033[1m'
    END = '\033[0m'

    """Pretty print the current result of the matches."""
    n_played = len(rr_played)
    es = "es" if n_played != 1 else ""
    n_togo = len(rr_unplayed)

    config.print()
    config.print('Ranking after {n_played} match{es} ({n_togo} to go):'.format(n_played=n_played, es=es, n_togo=n_togo))
    for team_id, p in round1_ranking(config, rr_played):
        if team_id in highlight:
            config.print("  {BOLD}{:>25}{END} {}".format(config.team_name(team_id), p, BOLD=BOLD, END=END))
        else:
            config.print("  {:>25} {}".format(config.team_name(team_id), p))
    config.print()


def play_round1(config, state):
    """Run the first round and return a sorted list of team names.

    teams is the sorted list [group0, group1, ...] and not the actual names of
    the agents. This is necessary to start the agents.
    """
    rr_unplayed = state.round1["unplayed"]
    rr_played = state.round1["played"]

    config.wait_for_keypress()
    config.print()
    config.print("ROUND 1 (Everybody vs Everybody)")
    config.print('================================', speak=False)
    config.print()


    match_id = MatchID(round=1)
    for played_match in rr_played:
        _logger.debug("Skipping played match {}.".format(played_match))
        match_id.next_match()
        t1_id, t2_id = played_match["match"]
        winner = played_match["winner"]
        if winner is False or winner is None:
            config.print("Already played match between {t1} and {t2}. (Draw.) Skipping.".format(t1=config.team_name_group(t1_id),
                                                                                                t2=config.team_name_group(t2_id)))
        else:
            config.print("Already played match between {t1} and {t2}. ({winner} won.) Skipping.".format(t1=config.team_name_group(t1_id),
                                                                                                        t2=config.team_name_group(t2_id),
                                                                                                        winner=config.team_name(winner)))

    if not rr_unplayed:
        pp_round1_results(config, rr_played, rr_unplayed)

    while rr_unplayed:
        match = rr_unplayed.pop()

        winner = start_match_with_replay(config, match, match_id=match_id)
        match_id.next_match()
        config.wait_for_keypress()

        if winner is False or winner is None:
            rr_played.append({ "match": match, "winner": False })
        else:
            rr_played.append({ "match": match, "winner": winner })

        pp_round1_results(config, rr_played, rr_unplayed, highlight=match)

        state.save(config.statefile)

    return [team_id for team_id, p in round1_ranking(config, rr_played)]


def recur_match_winner(match):
    """ Returns the team id of the unambiguous winner.

    Parameters
    ----------
    match : one of Match, Bye, Team or a team id (str or int)
        Match to get the id for

    Returns
    -------
    team_id or None :
        The config id of the team.

    """

    if isinstance(match, knockout_mode.Match) and match.winner is not None:
        return recur_match_winner(match.winner)
    elif isinstance(match, knockout_mode.Bye):
        return recur_match_winner(match.team)
    elif isinstance(match, knockout_mode.Team):
        return match.name
    elif isinstance(match, str) or isinstance(match, int):
        return match
    return None


def play_round2(config, teams, state):
    """Run the second round and return the name of the winning team.

    teams is the list [group0, group1, ...] not the names of the agens, sorted
    by the result of the first round.
    """
    config.print()
    config.print('ROUND 2 (K.O.)')
    config.print('==============', speak=False)
    config.print()
    config.wait_for_keypress()

    tournament = state.round2.get("tournament")
    last_match = state.round2.get("last_match")
    if tournament and last_match:
        config.print("Loading from state.", speak=False)
    else:
        last_match = knockout_mode.prepare_matches(teams, bonusmatch=config.bonusmatch)
        tournament = knockout_mode.tree_enumerate(last_match)

        state.round2["last_match"] = last_match
        state.round2["tournament"] = tournament

    config.print(knockout_mode.print_knockout(last_match, config.team_name), speak=False)

    match_id = MatchID(round=2, match=1)
    for round in tournament:
        for match in round:
            if isinstance(match, knockout_mode.Match):
                t1_id = recur_match_winner(match.t1)
                t2_id = recur_match_winner(match.t2)
                if not match.winner:
                    winner = start_deathmatch(config, t1_id, t2_id, match_id=match_id)
                    match.winner = winner

                    config.print(knockout_mode.print_knockout(last_match, config.team_name, highlight=[match]), speak=False)

                    state.round2["tournament"] = tournament
                    state.save(config.statefile)
                else:
                    _logger.debug("Skipping match {}.".format(match))
                    config.print("Already played match between {t1} and {t2}. ({winner} won.) Skipping.".format(t1=config.team_name_group(t1_id),
                                                                                                                t2=config.team_name_group(t2_id),
                                                                                                                winner=config.team_name(match.winner)))
                match_id.next_match()

    config.wait_for_keypress()

    return last_match.winner



#!/usr/bin/env python3

import contextlib
import hashlib
import importlib
import json
import logging
from pathlib import Path
import sys

import click
import zmq

from ..team import make_team
from ..network import json_default_handler
from .script_utils import start_logging

_logger = logging.getLogger(__name__)


@contextlib.contextmanager
def with_sys_path(dirname):
    sys.path.insert(0, dirname)
    try:
        yield
    finally:
        sys.path.remove(dirname)

def run_player(team_spec, address, team_name_override=False, silent_bots=False):
    """ Creates a team from `team_spec` and runs
    a game through the zmq PAIR socket on `address`.

    Parameters
    ----------
    team_spec : str
        path to the module that declares the team
    address : address to zmq PAIR socket
        the address of the remote team socket

    """

    address = address.replace('*', 'localhost')
    # Connect to the given address
    context = zmq.Context()
    socket = context.socket(zmq.PAIR)
    try:
        socket.connect(address)
    except zmq.ZMQError as e:
        raise IOError(f"Failed to connect the client to address {address}: {e}")

    try:
        team = load_team(team_spec)
    except Exception as e:
        # We could not load the team.
        # Wait for the set_initial message from the server
        # and reply with an error.
        try:
            json_message = socket.recv_unicode()
            py_obj = json.loads(json_message)
            uuid_ = py_obj["__uuid__"]
            action = py_obj["__action__"]
            data = py_obj["__data__"]

            socket.send_json({
                '__uuid__': uuid_,
                '__error__': e.__class__.__name__,
                '__error_msg__': f'Could not load {team_spec}: {e}'
            })
        except zmq.ZMQError as e:
            raise IOError('failed to connect the client to address %s: %s'
                          % (address, e))
        # TODO: Do not raise here but wait for zmq to return a sensible error message
        # We need a way to distinguish between syntax errors in the client
        # and general zmq disconnects
        raise

    _logger.info(f"Running player '{team_spec}' ({team.team_name})")

    while True:
        cont = player_handle_request(socket, team, team_name_override=team_name_override, silent_bots=silent_bots)
        if not cont:
            return


def player_handle_request(socket, team, team_name_override=False, silent_bots=False):
    """ Awaits a new request on `socket` and dispatches it
    to `team`.

    Parameters
    ----------
    socket : zmq PAIR socket
        the connection to the main pelita game
    team : a Team object
        the team that handles the requests

    Returns
    -------
    continue_processing : bool
        True if still running, False on exit

    """

    # Waits for incoming requests and tries to get a proper
    # answer from the player.

    try:
        json_message = socket.recv_unicode()
        py_obj = json.loads(json_message)
        msg_id = py_obj["__uuid__"]
        action = py_obj["__action__"]
        data = py_obj["__data__"]
        _logger.debug("<o-- %r [%s]", action, msg_id)

        # feed client actor here …
        if action == "set_initial":
            retval = team.set_initial(**data)
        elif action == "get_move":
            retval = team.get_move(**data)
            if silent_bots:
                # We want to remove a speak attribute
                # but we don’t care if it fails at all
                try:
                    retval.pop('say')
                except:
                    pass

        elif action == "team_name":
            if isinstance(team_name_override, str):
                retval = team_name_override
            else:
                retval = team.team_name
            # TODO: Log team name override
        elif action == "exit":
            # quit and don’t return anything
            message_obj = {
                "__uuid__": msg_id,
                "__return__": None
            }
            return False
        else:
            _logger.warning(f"Player received unknown action {action}.")

        message_obj = {
            "__uuid__": msg_id,
            "__return__": retval
        }

        # retval can be a string with a team name,
        # a dict with a move and optionally additional data,
        # or a dict with error key and message, if something
        # went wrong

        if isinstance(retval, dict) and 'error' in retval:
            # The team class has flagged an error.
            # We return the result (in the finally clause)
            # but return false to exit the process.
            return False
        else:
            # continue
            return True

    except KeyboardInterrupt as e:
        # catch KeyboardInterrupt to avoid spamming stderr
        msg_id = None
        message_obj = {
            '__error__': e.__class__.__name__
        }
        return True

    except Exception as e:
        # All client exceptions should have been caught in the
        # team class. This clause is a safety net for
        # exceptions that are not caused by a bot.

        msg = "Exception in client code for team %s." % team
        print(msg, file=sys.stderr)
        message_obj = {
            '__uuid__': msg_id,
            '__error__': e.__class__.__name__,
            '__error_msg__': str(e)
        }
        return False

    finally:
        # we use our own json_default_handler
        # to automatically convert numpy ints to json
        json_message = json.dumps(message_obj, default=json_default_handler)
        # return the message
        socket.send_unicode(json_message)
        if '__error__' in message_obj:
            _logger.warning("o-!> %r [%s]", message_obj['__error__'], msg_id)
        else:
            _logger.debug("o--> %r [%s]", message_obj['__return__'], msg_id)


def check_team_name(name):
    # Team name must be ascii
    try:
        name.encode('ascii')
    except UnicodeEncodeError:
        raise ValueError('Invalid team name (non ascii): "%s".'%name)
    # Team name must be shorter than 25 characters
    if len(name) > 25:
        raise ValueError('Invalid team name (longer than 25): "%s".'%name)
    if len(name) == 0:
        raise ValueError('Invalid team name (too short).')
    # Check every character and make sure it is either
    # a letter or a number. Nothing else is allowed.
    for char in name:
        if (not char.isalnum()) and (char != ' '):
            raise ValueError('Invalid team name (only alphanumeric '
                             'chars or blanks): "%s"'%name)
    if name.isspace():
        raise ValueError('Invalid team name (no letters): "%s"'%name)


def load_team(spec):
    """ Tries to load a team from a given spec.

    At first it will check if it can build a team from the built-in players.
    If this fails, it will try to load a factory.
    """
    if spec == '0':
        from ..player import SmartEatingPlayer
        return team_from_module(SmartEatingPlayer)
    elif spec == '1':
        from ..player import SmartRandomPlayer
        return team_from_module(SmartRandomPlayer)
    try:
        team = load_team_from_module(spec)
    except (FileNotFoundError, ImportError, ValueError,
            AttributeError, TypeError, IOError) as e:
        print("Failure while loading team '%s'" % spec, file=sys.stderr)
        print('ERROR: %s' % e, file=sys.stderr)
        raise

    check_team_name(team.team_name)
    return team

def load_team_from_module(path: str):
    """ Tries to load and return a team from a given path.

    The path should be a Python module (which can be a folder with an
    __init__.py file) or an importable .py file.
    The module must contain a function `move` and a variable `TEAM_NAME`.

    Parameters
    ----------
    pathspec : str
        the location of the importable module

    Returns
    -------
    factory : function
        the factory method that should produce a team

    Raises
    ------
    ValueError
        if a module is already present in sys.modules
    AttributeError
        if the module has no factory with the given name
    ModuleNotFoundError
        if the module cannot be found
    FileNotFoundError
        if the parent folder cannot be found
    """
    path = Path(path)

    if not path.parent.exists():
        raise FileNotFoundError("Folder {} does not exist.".format(path.parent))

    dirname = str(path.parent)
    modname = path.stem

    if modname in sys.modules:
        raise ValueError("A module named ‘{}’ has already been imported.".format(modname))

    with with_sys_path(dirname):
        module = importlib.import_module(modname)

    return team_from_module(module)


def team_from_module(module):
    """ Looks for a move function and a team name in
    `module` and returns a team.
    """
    # look for a new-style team
    move = module.move
    name = module.TEAM_NAME
    if not callable(move):
        raise TypeError("move is not a function")
    if not isinstance(name, str):
        raise TypeError("TEAM_NAME is not a string")
    team, _ = make_team(move, team_name=name)
    return team


@click.group()
@click.option('--log',
              is_flag=False, flag_value="-", default=None, metavar='LOGFILE',
              help="print debugging log information to LOGFILE (default 'stderr')")
def main(log):
    if log is not None:
        start_logging(log)


@main.command(help="Load team and connect to the specified address.")
@click.argument('team')
@click.argument('address')
@click.option('--team-name-override',
              default=None,
              help='Override the team name')
@click.option('--silent-bots',
              is_flag=True,
              default=False,
              help='Filter bot speak')
def remote_game(team, address, team_name_override, silent_bots):
    run_player(team, address, team_name_override=team_name_override, silent_bots=silent_bots)


@main.command("check-team", help="Load team and print its name.")
@click.argument('team')
def cli_check_team(team):
    return check_team(team)

@main.command("hash-team", help="Load team and print its hash.")
@click.argument('team')
def cli_hash_team(team):
    print(hash_team(team))

def check_team(team):
    print(load_team(team).team_name)

def hash_team(team):
    # Load the team so that we have the modules ready
    load_team(team)

    folder = Path(team).parent.resolve()
    modules = []
    for name, module in sys.modules.items():
        if hasattr(module, '__file__') and module.__file__:
            path = Path(module.__file__)
            if path.is_relative_to(folder):
                modules.append([name, path])

    _logger.debug(f"Hashing module {team}")

    # sort relative paths by module name and generate our sha
    sha1 = hashlib.sha1()
    for module, path in sorted(modules):
        _logger.debug(f"Hashing {team}: Adding {module}")
        sha1.update(path.read_bytes())
    res = sha1.hexdigest()
    _logger.debug(f"SHA1 for {team}: {res}.")
    return res


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Python script to generate a new maze.

The maze will be sent to sys.out .
"""

import sys
import argparse

from pelita.maze_generator import get_new_maze

HEIGHT = 16
WIDTH = 32
FOOD = 30
SEED = None

description="""
Return a new random layout to be used by pelita
"""
epilog=f"""
NOTE: To re-create the layouts used in the installed pelita, you can use the
following shell script:

# initialize random seed, any subsequent evaluation of $RANDOM will get
# a different, but replicable random number
RANDOM=39285

# keep track of the seeds
SEEDF="_seeds"
echo > $SEEDF
echo "## pelita-createlayout -y {HEIGHT} -x {WIDTH} -f {FOOD} -s SEED > normal_XXX.layout" >> $SEEDF
echo "## pelita-createlayout -y {HEIGHT//2} -x {WIDTH//2} -f {FOOD//3} -s SEED > small_XXX.layout" >> $SEEDF
echo "## pelita-createlayout -y {HEIGHT*2} -x {WIDTH*2} -f {FOOD*2} -s SEED > big_XXX.layout" >> $SEEDF

# generate 1000 normal layouts
for COUNT in $(seq -w 0 999); do
    echo "Generating normal_$COUNT..."
    SEED=$RANDOM$RANDOM$RANDOM
    pelita-createlayout -y {HEIGHT} -x {WIDTH} -f {FOOD} -s $SEED > normal_${{COUNT}}.layout
    echo "normal_${{COUNT}} = $SEED" >> $SEEDF
done

# generate 100 small layouts
for COUNT in $(seq -w 0 99); do
    echo "Generating small_0$COUNT..."
    SEED=$RANDOM$RANDOM$RANDOM
    pelita-createlayout -y {HEIGHT//2} -x {WIDTH//2} -f {FOOD//3} -s $SEED > small_0${{COUNT}}.layout
    echo "small_0${{COUNT}} = $SEED" >> $SEEDF
done

# generate 100 big layouts
for COUNT in $(seq -w 0 99); do
    echo "Generating big_0$COUNT..."
    SEED=$RANDOM$RANDOM$RANDOM
    pelita-createlayout -y {HEIGHT*2} -x {WIDTH*2} -f {FOOD*2} -s $SEED > big_0${{COUNT}}.layout
    echo "big_0${{COUNT}} = $SEED" >> $SEEDF
done
"""

def default(text, default):
    return text + f' (default: {default})'

def main():
    parser = argparse.ArgumentParser(
                  description=description,
                  epilog=epilog,
                  formatter_class=argparse.RawDescriptionHelpFormatter,
                  add_help=True)

    parser.add_argument('--height', '-y', default=HEIGHT, type=int, metavar='Y',
                        help=default('height of the maze', HEIGHT))
    parser.add_argument('--width', '-x', default=WIDTH, type=int, metavar='X',
                        help=default('width of the maze', WIDTH))
    parser.add_argument('--food', '-f', default=FOOD, type=int, metavar='F',
                        help=default('food pellets for each team', FOOD))
    parser.add_argument('--seed', '-s', default=SEED, type=int, metavar='S',
                        help=default('random seed', SEED))

    args = parser.parse_args()

    maze_str = get_new_maze(args.height, args.width, nfood=args.food,
                            seed=args.seed)

    sys.stdout.write(maze_str+'\n')
    sys.stdout.close()

if __name__ == "__main__":
    main()



import json
import logging
import os
import shutil
import sys

import tkinter
import zmq

from .tk_canvas import TkApplication

_logger = logging.getLogger(__name__)

def force_frontmost():
    """ This hack was discussed in https://github.com/ASPP/pelita/issues/345
    and uses Apple Script to tell OS X’s System Events to put a certain
    process in the foreground, focused and at the top position in the
    application stack. (I.e. cmd+tab behaves as expected.)
    """
    if sys.platform == 'darwin' and shutil.which('/usr/bin/osascript'):
        script = 'tell application "System Events" \
                  to set frontmost of the first process whose unix id is {pid} to true'.format(pid=os.getpid())
        os.system("/usr/bin/osascript -e '{script}'".format(script=script))


class TkViewer:
    """ Initialises Tk based viewer for the game.

    The viewer may be passed to a GameMaster instance by calling::

        viewer = TkViever(zmq_address)
        gm.register_viewer(viewer)

    Afterwards, Tk needs to run in the main thread, which is done
    by calling::

        viewer.run()

    Notes
    -----
    Any Tk application must run in the (or at least *a*) main thread.
    Therefore, the real game needs to run in some background thread, or
    the Tk application needs to have its own process. (Note that using
    Python’s subprocess library might have it’s very own issues with Tk.)

    This means that we’ll need to exchange all information about current
    universe states in a thread-safe and location independent way. A good way
    to accomplish this is using a zmq socket.

    There is however a problem with a simple message passing approach: The
    producer (ie. the GameMaster) may produce new states much faster than the
    consumer is able to process them. (ie. the Viewer might show some
    animations which just take their time.)
    A message queue which is just naïvely filled might therefore be a
    suboptimal solution when the game is finished long before the animations.

    This synchronisation could be accomplished using a separate messaging
    scheme which tells the main game that it should move on.

    Parameters
    ----------
    address : zmq uri
        The address of the zmq socket to connect to
    geometry: tuple, default = None
        The size (in pixel) of the game root window. None means
        using a bit less than the screen size.

    Attributes
    ----------
    root : The Tk root instance
    app : The TkApplication class

    """
    def __init__(self, address, controller_address=None, geometry=None, delay=1, stop_after=None, fullscreen=False):
        self.address = address
        self.controller_address = controller_address
        self.delay = delay
        self.geometry = geometry if geometry else (900, 580)
        self.fullscreen = fullscreen
        self.stop_after = stop_after

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.SUB)
        self.socket.setsockopt_unicode(zmq.SUBSCRIBE, "")
        self.socket.connect(self.address)
        self.poll = zmq.Poller()
        self.poll.register(self.socket, zmq.POLLIN)

        self._delay = 2

    def run(self):
        try:
            self.root = tkinter.Tk()
        except tkinter.TclError as e:
            _logger.error('TclError: %s. Exiting.', e)
            if self.controller_address:
                # We should be controlling the server. But we can’t.
                # Send exit.
                # TODO: This should flag an error code back to the server.
                context = zmq.Context()
                controller_socket = context.socket(zmq.DEALER)
                controller_socket.connect(self.controller_address)
                controller_socket.send_json({"__action__": "exit"})
            sys.exit(-1)
        if self.fullscreen:
            self.root.attributes('-fullscreen',True)
        else:
            root_geometry = str(self.geometry[0])+'x'+str(self.geometry[1])
            # put the root window in some sensible position
            self.root.geometry(root_geometry+'+40+40')

        self.app = TkApplication(window=self.root,
                                 controller_address=self.controller_address,
                                 geometry=self.geometry,
                                 delay=self.delay,
                                 stop_after=self.stop_after, fullscreen=self.fullscreen)
        # schedule next read
        self.root.after_idle(self.read_queue)
        try:
            # Try our best to get the application to the front.
            self.root.lift()
            force_frontmost()

            self.root.mainloop()
        except KeyboardInterrupt:
            pass

    def read_queue(self):
        # We increase the polling delay up to 100ms each try to avoid
        # using too much cpu when there are no messages anyway.
        if self._delay > 100:
            self._delay = 100
        try:
            # read all events.
            # if queue is empty, try again in a few ms
            # we don’t want to block here and lock
            # Tk animations
            message = self.socket.recv_unicode(flags=zmq.NOBLOCK)
            message = json.loads(message)

            _logger.debug(message["__action__"])
            # we currently don’t care about the action
            game_state = message["__data__"]
            if game_state:
                self.app.observe(game_state)

            self._delay = 2
            self._after(2, self.read_queue)
        except zmq.Again as e:
            _logger.debug('Nothing received. Waiting %.3d seconds.', self._delay)
            self._after(self._delay, self.read_queue)
            self._delay = self._delay * 2
        except zmq.ZMQError as e:
            _logger.info('ZMQ Error: %r. Ignoring.', e)
            self._after(self._delay, self.read_queue)
            self._delay = self._delay * 2

    def _after(self, delay, fun, *args):
        """ Execute fun(*args) after delay milliseconds.

        # Patched to quit after `KeyboardInterrupt`s.
        """
        def wrapped_fun():
            try:
                fun(*args)
            except KeyboardInterrupt:
                print("KBI")
                _logger.info("Detected KeyboardInterrupt. Exiting.")
                self.quit()
        self.root.after(delay, wrapped_fun)



import pytest
import unittest

import random
import time

from pelita.game import setup_game, run_game, play_turn
from pelita.layout import parse_layout
from pelita.player import (random_player, stopping_player, stepping_player,
                           round_based_player, speaking_player)


# some directions that are used in the tests
north = (0, -1)
south = (0, 1)
west  = (-1, 0)
east  = (1, 0)
stop  = (0, 0)

class TestSteppingPlayer:
    def test_stepping_players(self):
        test_layout = (
        """ ############
            #a  .  .  x#
            #b        y#
            ############ """)
        movements_0 = [east, east]
        movements_1 = [west, west]
        teams = [
            stepping_player(movements_0, movements_0),
            stepping_player(movements_1, movements_1)
        ]
        state = setup_game(teams, layout_dict=parse_layout(test_layout), max_rounds=2)
        assert state['bots'] == [(1, 1), (10, 1), (1, 2), (10, 2)]
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=2)
        assert state['bots'] == [(3, 1), (8, 1), (3, 2), (8, 2)]

    def test_shorthand(self):
        test_layout = (
        """ ############
            #a  .  .  y#
            #b        x#
            ############ """)
        num_rounds = 5
        teams = [
            stepping_player('>v<^-', '-----'),
            stepping_player('<^>v-', '-----')
        ]
        state = setup_game(teams, layout_dict=parse_layout(test_layout), max_rounds=5)
        player0_expected_positions = [(1,1), (2,1), (2,2), (1,2), (1,1), (1, 1)]
        player1_expected_positions = [(10,2), (9,2), (9,1), (10,1), (10,2), (10, 2)]

        assert state['bots'][0] == player0_expected_positions[0]
        assert state['bots'][1] == player1_expected_positions[0]
        for i in range(1, num_rounds+1):
            for step in range(4):
                state = play_turn(state)
            assert state['bots'][0] == player0_expected_positions[i]
            assert state['bots'][1] == player1_expected_positions[i]


    def test_too_many_moves(self):
        test_layout = (
        """ ############
            #a  .  .  x#
            #b        y#
            ############ """)
        movements_0 = [east, east]
        movements_1 = [west, west]
        teams = [
            stepping_player(movements_0, movements_0),
            stepping_player(movements_1, movements_1)
        ]
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=2)
        assert state['fatal_errors'] == [[], []]
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=3)
        assert len(state['fatal_errors'][0])
        # TODO: check for exact turn/round of the failure


class TestRoundBasedPlayer:
    def test_round_based_players(self):
        test_layout = (
        """ ############
            #a  .  .  x#
            #b        y#
            ############ """)
        # index 0 can be ignored
        movements_0 = [None, east, east]
        movements_1_0 = {1: west, 3: west}
        movements_1_1 = {3: west}
        teams = [
            round_based_player(movements_0, movements_0),
            round_based_player(movements_1_0, movements_1_1)
        ]
        state = setup_game(teams, layout_dict=parse_layout(test_layout), max_rounds=3)
        assert state['bots'][0] == (1, 1)
        assert state['bots'][1] == (10, 1)
        assert state['bots'][2] == (1, 2)
        assert state['bots'][3] == (10, 2)

        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=3)
        assert state['bots'][0] == (3, 1)
        assert state['bots'][1] == (8, 1)
        assert state['bots'][2] == (3, 2)
        assert state['bots'][3] == (9, 2)

class TestRandomPlayerSeeds:
    def test_demo_players(self):
        test_layout = (
        """ ################
            #b            y#
            #              #
            #              #
            #   a      x   #
            #              #
            #              #
            #              #
            #.            .#
            ################ """)
        teams = [
            random_player,
            random_player
        ]
        state = setup_game(teams, layout_dict=parse_layout(test_layout), max_rounds=20, seed=20)
        assert state['bots'][0] == (4, 4)
        assert state['bots'][1] == (4 + 7, 4)

        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=20, seed=20)
        pos_left_bot = state['bots'][0]
        pos_right_bot = state['bots'][1]

        # running again to test seed:
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=20, seed=20)
        assert state['bots'][0] == pos_left_bot
        assert state['bots'][1] == pos_right_bot

        # running again with other seed:
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=20, seed=200)
        # most probably, either the left bot or the right bot or both are at
        # a different position
        assert not (state['bots'][0] == pos_left_bot and state['bots'][1] == pos_right_bot)

    def test_random_seeds(self):
        test_layout = (
        """ ################
            #              #
            #              #
            #              #
            #   a      x   #
            #   b      y   #
            #              #
            #              #
            #.            .#
            ################ """)
        def init_rng_players():
            player_rngs = []
            def rng_test(bot, state):
                player_rngs.append(bot.random)
                return bot.position, state
            team = [
                rng_test,
                rng_test
            ]
            return team, player_rngs

        team0, player_rngs0 = init_rng_players()
        state = setup_game(team0, layout_dict=parse_layout(test_layout), max_rounds=5, seed=20)
        # play two steps
        play_turn(play_turn(state))
        assert len(player_rngs0) == 2
        # generate some random numbers for each player
        random_numbers0 = [rng.randint(0, 10000) for rng in player_rngs0]
        # teams should have generated a different number
        assert random_numbers0[0] != random_numbers0[1]

        team1, player_rngs1 = init_rng_players()
        state = setup_game(team1, layout_dict=parse_layout(test_layout), max_rounds=5, seed=20)
        # play two steps
        play_turn(play_turn(state))
        assert len(player_rngs1) == 2
        # generate some random numbers for each player
        random_numbers1 = [rng.randint(0, 10000) for rng in player_rngs1]
        # teams should have generated the same numbers as before
        assert random_numbers0 == random_numbers1

        # now, use a different seed
        team2, player_rngs2 = init_rng_players()
        state = setup_game(team2, layout_dict=parse_layout(test_layout), max_rounds=5, seed=200)
        # play two steps
        play_turn(play_turn(state))
        assert len(player_rngs2) == 2
        # generate some random numbers for each player
        random_numbers2 = [rng.randint(0, 10000) for rng in player_rngs0]
        # teams should have generated different numbers than before
        assert random_numbers0 != random_numbers2


def test_speaking_player():
    test_layout = (
    """ ############
        #ab#.  .#yx#
        ############ """)
    teams = [
        speaking_player,
        random_player
    ]
    state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=1)
    assert state["say"][0].startswith("Going")
    assert state["say"][1] == ""




import cmath
import math
from .. import layout
from ..gamestate_filters import manhattan_dist


def col(red, green, blue):
    """Convert the given colours [0, 255] to HTML hex colours."""
    return "#%02x%02x%02x" % (red, green, blue)

RED = col(235, 90, 90)
BLUE = col(94, 158, 217)

LIGHT_BLUE = '#B9D9F6'
STRONG_BLUE = '#1E6BB1'
LIGHT_RED = '#FFB0B0'
STRONG_RED = '#A91919'

YELLOW = col(242, 255, 83)
YELLOW = '#FFE38B'
GREY = col(80, 80, 80)
LIGHT_GREY = col(230, 230, 230)
SELECTED = col(200, 200, 200)
BROWN = col(48, 26, 22)

SHADOW_RED = '#B37373'
SHADOW_BLUE = '#6D92B3'

FOOD_WARNING_TIME = 6

def rotate(arc, rotation):
    """Helper for rotation normalisation."""
    return (arc + rotation) % 360

def pos_to_complex(pos):
    x, y = pos
    return x - y * 1j

class TkSprite:
    def __init__(self, mesh, *, position=None, _tag=None):
        self.mesh = mesh

        self._position = position
        self._direction = None
        self._tag = _tag

    @property
    def direction(self):
        return self._direction

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        if self.position is None or position is None:
            self._direction = None
            self._position = position
            return

        old_pos = self._position
        new_pos = position

        self._position = new_pos

        # automatic rotation
        if new_pos != old_pos:
            self._direction = math.degrees(cmath.phase(pos_to_complex(new_pos) - pos_to_complex(old_pos)))

    def screen(self, shift=(0, 0)):
        x, y = self.position
        return self.mesh.mesh_trafo(x, y).screen(*shift)

    def draw(self, canvas, game_state=None):
        raise NotImplementedError

    def bounding_box(self, scale_factor=1.0):
        return (self.screen((- scale_factor, - scale_factor)),
                self.screen((+ scale_factor, + scale_factor)))

    @property
    def tag(self):
        _tag = self._tag or "tag" + str(id(self))
        return _tag

    def redraw(self, canvas, game_state=None):
        self.delete(canvas)
        self.draw(canvas, game_state)

    def delete(self, canvas):
        canvas.delete(self.tag)

class BotSprite(TkSprite):
    def __init__(self, mesh, team=0, bot_id=0, shadow=False, **kwargs):
        self.bot_id = bot_id
        self.team = team
        self.width = mesh.mesh_width

        self.shadow = shadow

        if self.team == 0:
            self.col = BLUE if not self.shadow else ""
            self.outline_col = BLUE if not self.shadow else SHADOW_BLUE
            self.eye_col = YELLOW if not self.shadow else BLUE
        else:
            self.col = RED if not self.shadow else ""
            self.outline_col = RED if not self.shadow else SHADOW_RED
            self.eye_col = YELLOW if not self.shadow else RED

        self.is_harvester = None

        super(BotSprite, self).__init__(mesh, **kwargs)

    def is_harvester_at(self, pos):
        if self.team == 0:
            return pos[0] >= self.width // 2
        elif self.team == 1:
            return pos[0] < self.width // 2

    def delete(self, canvas):
        canvas.delete("speak" + self.tag)
        canvas.delete("show_id" + self.tag)
        super().delete(canvas)

    def move_to(self, new_pos, canvas, game_state=None, force=None, say="", show_id=False):
        old_direction = self.direction
        old_position = self.position

        self.position = new_pos
        if (old_position is None
            or old_direction != self.direction
            or force
            or self.is_harvester != self.is_harvester_at(game_state['bots'][self.bot_id])
            ):
            # We cannot just move the shape in these cases
            # We must redraw the whole shape
            self.redraw(canvas, game_state)
        else:
            dx = self.position[0] - old_position[0]
            dy = self.position[1] - old_position[1]

            canvas.move(self.tag, self.mesh.rect_width * dx, self.mesh.rect_height * dy)

        canvas.delete("speak"+self.tag)
        # We increase readability with a white border around the text.
        canvas.create_text(self.bounding_box()[0][0]-1, self.bounding_box()[0][1], text=say, font=(None, 12), fill="white", tag="speak"+self.tag)
        canvas.create_text(self.bounding_box()[0][0]+1, self.bounding_box()[0][1], text=say, font=(None, 12), fill="white", tag="speak"+self.tag)
        canvas.create_text(self.bounding_box()[0][0], self.bounding_box()[0][1]-1, text=say, font=(None, 12), fill="white", tag="speak"+self.tag)
        canvas.create_text(self.bounding_box()[0][0], self.bounding_box()[0][1]+1, text=say, font=(None, 12), fill="white", tag="speak"+self.tag)
        canvas.create_text(self.bounding_box()[0][0], self.bounding_box()[0][1], text=say, font=(None, 12), fill="black", tag="speak"+self.tag)

        canvas.delete("show_id" + self.tag)
        # we print the bot_id in the lower left corner
        if show_id:
            bot_name = layout.BOT_I2N[self.bot_id ]
            shift_x = 5
            shift_y = 8
            canvas.create_text(self.bounding_box()[0][0]-1 + shift_x, self.bounding_box()[1][1] - shift_y, text=bot_name, font=(None, 12), fill="white", tag="show_id"+self.tag)
            canvas.create_text(self.bounding_box()[0][0]+1 + shift_x, self.bounding_box()[1][1] - shift_y, text=bot_name, font=(None, 12), fill="white", tag="show_id"+self.tag)
            canvas.create_text(self.bounding_box()[0][0] + shift_x, self.bounding_box()[1][1]-1 - shift_y, text=bot_name, font=(None, 12), fill="white", tag="show_id"+self.tag)
            canvas.create_text(self.bounding_box()[0][0] + shift_x, self.bounding_box()[1][1]+1 - shift_y, text=bot_name, font=(None, 12), fill="white", tag="show_id"+self.tag)
            canvas.create_text(self.bounding_box()[0][0] + shift_x, self.bounding_box()[1][1] - shift_y, text=bot_name, font=(None, 12), fill="black", tag="show_id"+self.tag)

    def draw_bot(self, canvas, is_blue=True):
        direction = self.direction
        # set default direction, if we start from our initial position
        # If we are a shadow, stay fixed.
        if direction is None or self.shadow:
            direction = 0 if is_blue else 180

        # ensure that our eyes are never on the bottom
        if direction == 0:
            flip = True
        else:
            flip = False

        # bot body
        canvas.create_arc(self.bounding_box(), start=rotate(20, direction), extent=320, style="pieslice",
                          width=0, outline=self.outline_col, fill=self.col, tag=self.tag)

        # bot eye
        # first locate eye in the center
        eye_size = 0.15
        eye_box = (-eye_size - eye_size * 1j, eye_size + eye_size * 1j)
        # shift it to the middle of the bot just over the mouth
        eye_box = [item + 0.4 + 0.6j for item in eye_box]
        # take also care of flipping
        if flip:
            eye_box = [item.conjugate() for item in eye_box]
        # rotate based on direction
        eye_box = [cmath.exp(1j * math.radians(-direction)) * item for item in eye_box]
        eye_box = [self.screen((item.real, item.imag)) for item in eye_box]
        canvas.create_oval(eye_box, fill=self.eye_col, width=0, tag=self.tag)

    def draw(self, canvas, game_state):
        self.is_harvester = self.is_harvester_at(self.position)

        if self.is_harvester:
            if self.team == 0:
                self.draw_bot(canvas, is_blue=True)
            else:
                self.draw_bot(canvas, is_blue=False)
        else:
            if self.team == 0:
                self.draw_destroyer(canvas)
            else:
                self.draw_destroyer(canvas)

    def draw_destroyer(self, canvas):
        box_ll, box_tr = self.bounding_box()

        # ghost head
        canvas.create_arc((box_ll, box_tr), start=0, extent=180, style="pieslice" if not self.shadow else "arc",
                          width=1, outline=self.outline_col, fill=self.col, tag=self.tag)

        # ghost body
        box_ll = box_ll[0], box_ll[1] + (box_tr[1]-box_ll[1])/2.
        amplitude = (box_tr[1]-box_ll[1])/4.
        start = box_ll[0], box_tr[1]-amplitude/2.
        end = box_tr[0], start[1]
        periods = 3
        period = (end[0]-start[0])/periods
        points_per_period = 10
        x_spacing = period/points_per_period
        x = [start[0]+i*x_spacing for i in range(periods*points_per_period)]
        x.append(end[0])
        if period == 0:
            y = [start[1] for x_i in x]
        else:
            y = [amplitude*math.cos(2*math.pi*(x_i-start[0])/period) + start[1] for x_i in x]
        # add container edges for the polygon
        x.insert(0, box_ll[0]); y.insert(0, box_ll[1] - 1)
        x.append(box_tr[0]); y.append(box_ll[1] - 1)

        points = list(zip(x,y))
        if self.shadow:
            canvas.create_line(points, width=1, fill=self.outline_col, tag=self.tag)
        else:
            canvas.create_polygon(points, width=1, outline=self.outline_col, fill=self.col, tag=self.tag)

        # ghost eyes
        eye_size = 0.15
        eye_box = (-eye_size -eye_size*1j, eye_size + eye_size*1j)
        # right eye
        eye_box_r = [item+ 0.4 - 0.5j for item in eye_box]
        eye_box_r = [self.screen((item.real, item.imag)) for item in eye_box_r]
        canvas.create_oval(eye_box_r, fill=self.eye_col, width=0, tag=self.tag)
        # left eye
        eye_box_l = [item- 0.4 - 0.5j for item in eye_box]
        eye_box_l = [self.screen((item.real, item.imag)) for item in eye_box_l]
        canvas.create_oval(eye_box_l, fill=self.eye_col, width=0, tag=self.tag)

class Wall(TkSprite):
    def __init__(self, mesh, wall_neighbors=None, **kwargs):
        if wall_neighbors is None:
            self.wall_neighbors = []
        else:
            self.wall_neighbors = wall_neighbors

        super(Wall, self).__init__(mesh, **kwargs)


    def draw(self, canvas, game_state=None):
        scale = (self.mesh.half_scale_x + self.mesh.half_scale_y) * 0.6
        if not ((0, 1) in self.wall_neighbors or
                (1, 0) in self.wall_neighbors or
                (0, -1) in self.wall_neighbors or
                (-1, 0) in self.wall_neighbors):
            # if there is no direct neighbour, we can’t connect.
            # draw only a small dot.
            # TODO add diagonal lines
            canvas.create_line(self.screen((-0.3, 0)), self.screen((+0.3, 0)), fill=BROWN,
                               width=scale, tag=(self.tag, "wall"), capstyle="round")
        else:
            neighbours = [(-1, -1), (0, -1), (1, -1), (1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0)]
            for dx in [-1, 0, 1]:
                for dy in [-1, 0, 1]:
                    if (dx, dy) in self.wall_neighbors:
                        if dx == dy == 0:
                            continue
                        if dx * dy != 0:
                            continue
                        index = neighbours.index((dx, dy))
                        if (neighbours[(index + 1) % len(neighbours)] in self.wall_neighbors and
                            neighbours[(index - 1) % len(neighbours)] in self.wall_neighbors):
                            pass
                        else:
                            canvas.create_line(self.screen((0, 0)), self.screen((2*dx, 2*dy)), fill=BROWN,
                                               width=scale, tag=(self.tag, "wall"), capstyle="round")

            # if we are drawing a closed square, fill in the internal part
            # detect the square when we are on the bottom-left vertex of it
            square_neighbors = {(0,0), (0,-1), (1,0),(1,-1)}
            if square_neighbors <= set(self.wall_neighbors):
                canvas.create_line(self.screen((1,0)), self.screen((1,-2)), fill=BROWN,
                                   width=scale, tag=(self.tag, "wall"))


class Food(TkSprite):
    def __init__(self, mesh, food_age=None, max_food_age=None, **kwargs):
        self.food_age = food_age
        if max_food_age is None:
            self.max_food_age = math.inf
        else:
            self.max_food_age = max_food_age
        super().__init__(mesh, **kwargs)

    @classmethod
    def food_pos_tag(cls, position):
        return "Food" + str(position)

    def draw(self, canvas, game_state=None, show_lifetime=False):
        if self.position[0] < self.mesh.num_x/2:
            fill_col = BLUE
        else:
            fill_col = RED
        text_col = "#000"

        food_age = self.food_age

        if food_age and food_age + FOOD_WARNING_TIME > self.max_food_age:
            fill_col = GREY
            text_col = YELLOW
        canvas.create_oval(self.bounding_box(0.4), fill=fill_col, width=0, tag=(self.tag, self.food_pos_tag(self.position), "food"))

        canvas.delete("show_food_age" + str(self.position))

        # we print the bot_id in the lower left corner
        if food_age and show_lifetime:
            tag=(self.tag, "show_food_age" + str(self.position), "food")

            center = self.screen()
            canvas.create_text(*center, text=food_age, font=(None, 10), fill=text_col, tag=tag)

class Arrow(TkSprite):
    def __init__(self, mesh, req_pos, success, head=True, **kwargs):
        self.req_pos = req_pos
        self.success = success
        self.head = head

        super(Arrow, self).__init__(mesh, **kwargs)

    def draw(self, canvas, game_state=None):
        scale = (self.mesh.half_scale_x + self.mesh.half_scale_y) * 0.07

        if not self.success:
            points = [
                self.screen((-0.3, 0.3)),
                self.screen((0.3, -0.3))
            ]
            canvas.create_line(points,
                            fill=BROWN, width=scale, tag=(self.tag, "arrow"), capstyle="round")

            points = [
                self.screen((-0.3, -0.3)),
                self.screen(( 0.3, 0.3))
            ]
            canvas.create_line(points,
                            fill=BROWN, width=scale, tag=(self.tag, "arrow"), capstyle="round")

        dist = manhattan_dist(self.req_pos, self.position)
        if dist == 0:
            canvas.create_arc(self.bounding_box(0.6), start=110, extent=320, style="arc", outline=BROWN,
                                                width=scale, tag=(self.tag, "arrow"))
            # arrow head
            head = cmath.rect(0.6, -110 * cmath.pi / 180)
            head_rotation = (-110 + 90) * cmath.pi / 180
            head_left = head - cmath.rect(0.3, head_rotation - cmath.pi/4)
            head_right = head - cmath.rect(0.3, head_rotation + cmath.pi/4)
            #vector = dx + dy * 1j
            #phase = cmath.phase(vector)
            #head = vector + cmath.rect(0.1, phase)
            #head_left = vector - cmath.rect(1, phase) + cmath.rect(0.9, phase - cmath.pi/4)
            #head_right = vector - cmath.rect(1, phase) + cmath.rect(0.9, phase + cmath.pi/4)

            points = [
                self.screen((head_left.real, head_left.imag)),
                self.screen((head.real, head.imag)),
                self.screen((head_right.real, head_right.imag))
            ]
            canvas.create_line(points,
                            fill=BROWN, width=scale, tag=(self.tag, "arrow"), capstyle="round")

        else:
            # dx, dy has to be duplicated because the self.screen coordinates go from -1 to 1
            # for the current cell
            dx = (self.req_pos[0] - self.position[0]) * 2
            dy = (self.req_pos[1] - self.position[1]) * 2
            canvas.create_line(self.screen((0, 0)), self.screen((dx, dy)), fill=BROWN,
                                                width=scale, tag=(self.tag, "arrow"), capstyle="round")
            # arrow head
            vector = dx + dy * 1j
            phase = cmath.phase(vector)
            head = vector + cmath.rect(0.1, phase)
            head_left = vector - cmath.rect(1, phase) + cmath.rect(0.9, phase - cmath.pi/4)
            head_right = vector - cmath.rect(1, phase) + cmath.rect(0.9, phase + cmath.pi/4)

            points = [
                self.screen((head_left.real, head_left.imag)),
                self.screen((head.real, head.imag)),
                self.screen((head_right.real, head_right.imag))
            ]
            if self.head:
                canvas.create_line(points,
                                fill=BROWN, width=scale, tag=(self.tag, "arrow"), capstyle="round")




import importlib.resources as importlib_resources
import io
import random

# bot to index conversion
BOT_N2I = {'a': 0, 'b': 2, 'x': 1, 'y': 3}
BOT_I2N = {0: 'a', 2: 'b', 1: 'x', 3: 'y'}


RNG = random.Random()

def get_random_layout(size='normal', seed=None):
    """ Return a random layout string from the available ones.

    Parameters
    ----------
    size : str
        only return layouts of size 'small', 'normal', 'big', 'all'.
        Default is 'normal'.

        'small'  -> width=16, height=8,  food=10
        'normal' -> width=32, height=16, food=30
        'big'    -> width=64, height=32, food=60
        'all'    -> all of the above

    Returns
    -------
    layout : tuple(str, str)
        the name of the layout, a random layout string

    """
    if seed is not None:
        RNG.seed(seed)
    layouts_names = get_available_layouts(size=size)
    layout_choice = RNG.choice(layouts_names)
    return layout_choice, get_layout_by_name(layout_choice)

def get_available_layouts(size='normal'):
    """Return the names of the built-in layouts.

    Parameters
    ----------
    size : str
        only return layouts of size 'small', 'normal', 'big', 'all'.
        Default is 'normal'.

        'small'  -> width=16, height=8,  food=10
        'normal' -> width=32, height=16, food=30
        'big'    -> width=64, height=32, food=60
        'all'    -> all of the above


    Returns
    -------
    layout_names : list of str
        the available layouts

    """
    # loop in layouts directory and look for layout files
    valid = ('small', 'normal', 'big', 'all')
    if size not in valid:
        raise ValueError(f"Invalid layout size ('{size}' given). Valid: {valid}")
    if size == 'all':
        size = ''

    av_layouts = []

    for resource in importlib_resources.files('pelita._layouts').iterdir():
        if resource.is_file() and resource.name.endswith('.layout') and size in resource.name:
            layout_name = resource.name.removesuffix('.layout')
            av_layouts.append(layout_name)

    return sorted(av_layouts)

def get_layout_by_name(layout_name):
    """Get a built-in layout by name

    Parameters
    ----------
    layout_name : str
        a valid layout name

    Returns
    -------
    layout_str : str
        the layout as a string

    Raises
    ------
    KeyError
        if the layout_name is not known

    See Also
    --------
    get_available_layouts
    """
    try:
        return importlib_resources.files('pelita._layouts').joinpath(layout_name + '.layout').read_text()
    except FileNotFoundError:
        # This happens if layout_name is not found in the layouts directory
        # reraise as ValueError with appropriate error message.
        raise ValueError(f"Layout: '{layout_name}' is not known.") from None

def parse_layout(layout_str, food=None, bots=None):
    """Parse a layout string.

    A valid layout string is enclosed by walls and rectangular:

     ########
     #a  .  #
     #b ## x#
     #  .  y#
     ########


    Return a dict
        {'walls': set_of_wall_coordinates,
         'food' : list_of_food_coordinates,
         'bots'  : list_of_bot_coordinates in the order (a,x,b,y),
         'shape': tuple of (height, width) of the layout}

    In the example above:
    {'walls': {(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (1, 0), (1, 4), (2, 0),
               (2, 4), (3, 0), (3, 2), (3, 4), (4, 0), (4, 2), (4, 4), (5, 0),
               (5, 4), (6, 0), (6, 4), (7, 0), (7, 1), (7, 2), (7, 3), (7, 4)},
     'food': [(3, 3), (4, 1)],
     'bots': [(1, 1), (6, 2), (1, 2), (6, 3)],
     'shape': (8, 4)}

    Additional food and bots can be passed:

      - food: a list of coordinates of additional food pellets
      - bots: a dictionary { char : (coord_x, coord_y)}, where char in 'a', 'b', 'x', 'y'
              bots specified this way override the coordinates of bots found in the string
    """

    if bots is None:
        bots = {}
    if food is None:
        food = []

    # set empty default values
    lwalls = set()
    lfood = []
    lbots = [None] * 4

    width = None
    # list of layout rows
    rows = []
    start = False
    for i, line in enumerate(layout_str.splitlines()):
        row = line.strip()
        if not row:
            # always ignore empty lines
            continue
        # a layout is always started by a full row of walls
        if not start:
            if row.count('#') != len(row):
                raise ValueError(f"Layout must be enclosed by walls (line: {i})!")
            else:
                # start the layout parsing
                start = True
                # set width of layout
                width = len(row)
                # check that width is even
                if width % 2:
                    raise ValueError(f"Layout width must be even (found {width})!")
                rows.append(row)
                continue
        # Here we are within the layout
        # every row must have the same length
        if len(row) != width:
            raise ValueError(f"Layout rows have differing widths (line: {i})!")
        # rows are always enclosed by walls
        if row[0] != '#' or row[-1] != '#':
            raise ValueError(f"Layout must be enclosed by walls (line:{i})!")
        # append current row to the list of rows
        rows.append(row)
        # detect closing row and ignore whatever follows
        if row.count('#') == len(row):
            start = False
            break

    if start:
        # layout has not been closed!
        raise ValueError(f"Layout must be enclosed by walls (line:{i})!")

    # height of the layout (y-axis)
    height = len(rows)

    # iterate through the grid of characters
    for y, row in enumerate(rows):
        for x, char in enumerate(row):
            coord = (x, y)
            # assign the char to the corresponding list
            if char == '#':
                # wall
                lwalls.add(coord)
            elif char == '.':
                # food
                lfood.append(coord)
            elif char == ' ':
                # empty
                continue
            elif char in BOT_N2I.keys():
                # legal bots
                bot_idx = BOT_N2I[char]
                if lbots[bot_idx] is not None:
                    # bot_idx has already been set before
                    raise ValueError(f"Cannot set bot {BOT_I2N[bot_idx]} to {coord} (already at {lbots[bot_idx]}).")
                lbots[bot_idx] = coord
            else:
                raise ValueError(f"Unknown character {char} in maze at {coord}!")
    missing_bots = []
    for i, bot in enumerate(lbots):
        if bot is None and BOT_I2N[i] not in bots:
            missing_bots.append(BOT_I2N[i])
    if missing_bots:
            raise ValueError(f"Missing bot(s): {missing_bots}")
    lfood.sort()

    # if additional food was supplied, we add it
    for c in food:
        if not ((0 <= c[0] < width) and (0 <= c[1] < height)):
            raise ValueError(f"food item at {c} is outside of maze!")
        elif c in lwalls:
            raise ValueError(f"food item at {c} is on a wall!")
        else:
            lfood = sorted(list(set(food + lfood)))

    if not (set(bots.keys()) <= set(BOT_N2I.keys())):
        raise ValueError(f"Invalid Bot names in {bots}.")

    # check if additional bots are on legal positions
    for bn, bpos in bots.items():
        if not ((0 <= bpos[0] < width) and (0 <= bpos[1] < height)):
            raise ValueError(f"bot {bn} at {bpos} is outside of maze!")
        elif bpos in lwalls:
            raise ValueError(f"bot {bn} at {bpos} is on a wall!")
        else:
            # override bots
            lbots[BOT_N2I[bn]] = bpos

    # build parsed layout, ensuring walls and food are sorted
    parsed_layout = {
        'walls': tuple(sorted(lwalls)),
        'food': sorted(lfood),
        'bots': lbots,
        'shape': (width, height)
    }

    return parsed_layout


def layout_as_str(*, walls, food=None, bots=None, shape=None):
    """Given a dictionary with walls, food and bots coordinates return a string layout representation

    Example:

    Given:
    {'walls': {(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (1, 0), (1, 4), (2, 0),
               (2, 4), (3, 0), (3, 2), (3, 4), (4, 0), (4, 2), (4, 4), (5, 0),
               (5, 4), (6, 0), (6, 4), (7, 0), (7, 1), (7, 2), (7, 3), (7, 4)},
     'food': [(3, 3), (4, 1)],
     'bots': [(1, 1), (6, 2), (1, 2), (6, 3)],
     'shape': (8, 4)}

    Return:
    ########
    #a  .  #
    #b ## x#
    #  .  y#
    ########

    Overlapping items are discarded. When overlapping, walls take precedence over
    bots, which take precedence over food.

    The shape is optional. When it does not match the borders of the maze, a ValueError
    is raised.
    """
    width, height = wall_dimensions(walls)

    if shape is not None and not (width, height) == shape:
        raise ValueError(f"Given shape {shape} does not match width and height of layout {(width, height)}.")

    # initialized empty containers
    if food is None:
        food = []
    if bots is None:
        bots = []

    out = io.StringIO()
    for y in range(height):
        for x in range(width):
            out_char = " "
            if (x, y) in walls:
                # always print walls
                out_char = '#'
            elif (x, y) in bots:
                # Bot has the next priority
                bot_ix = bots.index((x,y))
                out_char = BOT_I2N[bot_ix]
            elif (x, y) in food:
                out_char = '.'
             # close the row
            out.write(out_char)
        out.write('\n')
    return out.getvalue()


def wall_dimensions(walls):
    """ Given a list of walls, returns the shape of the maze as a tuple of (width, height)"""
    max_elem = max(walls)
    width = max_elem[0] + 1
    height = max_elem[1] + 1
    return (width, height)


def initial_positions(walls, shape):
    """Calculate initial positions.

    Given the list of walls, returns the free positions that are closest to the
    bottom left and top right corner. The algorithm starts searching from
    (1, height-2) and (width-2, 1) respectively and uses the Manhattan distance
    for judging what is closest. On equal distances, a smaller distance in the
    x value is preferred.
    """
    width, height = shape

    left_start = (1, height - 2)
    left = []
    right_start = (width - 2, 1)
    right = []

    dist = 0
    while len(left) < 2:
        # iterate through all possible x distances (inclusive)
        for x_dist in range(dist + 1):
            y_dist = dist - x_dist
            pos = (left_start[0] + x_dist, left_start[1] - y_dist)
            # if both coordinates are out of bounds, we stop
            if not (0 <= pos[0] < width) and not (0 <= pos[1] < height):
                raise ValueError("Not enough free initial positions.")
            # if one coordinate is out of bounds, we just continue
            if not (0 <= pos[0] < width) or not (0 <= pos[1] < height):
                continue
            # check if the new value is free
            if pos not in walls:
                left.append(pos)

            if len(left) == 2:
                break

        dist += 1

    dist = 0
    while len(right) < 2:
        # iterate through all possible x distances (inclusive)
        for x_dist in range(dist + 1):
            y_dist = dist - x_dist
            pos = (right_start[0] - x_dist, right_start[1] + y_dist)
            # if both coordinates are out of bounds, we stop
            if not (0 <= pos[0] < width) and not (0 <= pos[1] < height):
                raise ValueError("Not enough free initial positions.")
            # if one coordinate is out of bounds, we just continue
            if not (0 <= pos[0] < width) or not (0 <= pos[1] < height):
                continue
            # check if the new value is free
            if pos not in walls:
                right.append(pos)

            if len(right) == 2:
                break

        dist += 1

    # lower indices start further away
    left.reverse()
    right.reverse()
    return [left[0], right[0], left[1], right[1]]


def get_legal_positions(walls, shape, bot_position):
    """ Returns all legal positions that a bot at `bot_position`
    can go to.

     Parameters
    ----------
    walls : set of (int, int)
        position of the walls of current layout.
    bot_position: tuple
        position of current bot.

    Returns
    -------
    list
        legal positions

    Raises
    ------
    ValueError
        if bot_position invalid or on wall
    """
    width, height = shape
    if not (0, 0) <= bot_position < (width, height):
        raise ValueError(f"Position {bot_position} not inside maze ({width}x{height}).")
    if bot_position in walls:
        raise ValueError(f"Position {bot_position} is on a wall.")
    north = (0, -1)
    south = (0, 1)
    east = (1, 0)
    west = (-1, 0)
    stop = (0, 0)
    directions = [north, east, west, south, stop]
    potential_moves = [(i[0] + bot_position[0], i[1] + bot_position[1]) for i in directions]
    possible_moves = [i for i in potential_moves if i not in walls]
    return possible_moves



import pytest

from pathlib import Path
import random
import sys
import tempfile

import pelita.game
from pelita.tournament import call_pelita, run_and_terminate_process

_mswindows = (sys.platform == "win32")

FIXTURE_DIR = Path(__file__).parent.resolve() / 'fixtures'


# Runs the processes for the remote teams
# and sets up `remote_teams` as a pytest fixture
# The processes should automatically terminate then
# also, listens on different ports each time, to avoid collisions and dead
# locks
RNG = random.Random()

@pytest.fixture(scope="module")
def remote_teams():
    # get random port numbers within the range of dynamic ports
    port_stopping = RNG.randint(49153,65534)
    port_food_eater = port_stopping+1

    addr_stopping = "127.0.0.1"
    addr_food_eater = "127.0.0.1"
    remote = [sys.executable, '-m', 'pelita.scripts.pelita_server', 'remote-server', '--address', '127.0.0.1']

    remote_stopping = remote + ['--port', str(port_stopping), '--team', 'pelita/player/StoppingPlayer']
    remote_food_eater = remote + ['--port', str(port_food_eater), '--team', 'pelita/player/FoodEatingPlayer']

    teams = [f'pelita://127.0.0.1:{port_stopping}/Stopping_Players', f'pelita://127.0.0.1:{port_food_eater}/Food_Eating_Players']
    with run_and_terminate_process(remote_stopping):
        with run_and_terminate_process(remote_food_eater):
            yield teams


def test_remote_call_pelita(remote_teams):
    res, stdout, stderr = call_pelita(remote_teams, rounds=30, size='small', viewer='null', seed='2')
    assert res['whowins'] == 1
    assert res['fatal_errors'] == [[], []]
    # errors for call_pelita only contains the last thrown error, hence None
    # TODO: should be aligned so that call_pelita and run_game return the same thing
    assert res['errors'] == [None, None]


def test_remote_run_game(remote_teams):
    # TODO: A failure here freezes pytest
    layout = """
        ##########
        #  b  y  #
        #a  ..  x#
        ##########
        """
    state = pelita.game.run_game(remote_teams, max_rounds=30, layout_dict=pelita.layout.parse_layout(layout))
    assert state['whowins'] == 1
    assert state['fatal_errors'] == [[], []]
    assert state['errors'] == [{}, {}]


@pytest.mark.xfail(reason="TODO: Fails in CI for macOS. Unclear why.")
def test_remote_timeout():
    # We have a slow player that also generates a bad move
    # in its second turn.
    # We need to detect both.
    # To avoid timing issues, the blue player will also need to be a bit slower

    layout = """
        ##########
        #  b  y  #
        #a  ..  x#
        ##########
        """

    blue = FIXTURE_DIR / 'remote_timeout_blue.py'
    red = FIXTURE_DIR / 'remote_timeout_red.py'

    state = pelita.game.run_game([str(blue), str(red)],
                                 max_rounds=8,
                                 layout_dict=pelita.layout.parse_layout(layout),
                                 timeout_length=0.4)

    assert state['whowins'] == 0
    assert state['fatal_errors'] == [[], []]
    assert state['errors'] == [{},
        {(1, 1): {'description': '', 'type': 'PlayerTimeout'},
        (1, 3): {'bot_position': (-2, 0), 'reason': 'illegal move'},
        (2, 1): {'description': '', 'type': 'PlayerTimeout'},
        (2, 3): {'description': '', 'type': 'PlayerTimeout'},
        (3, 1): {'description': '', 'type': 'PlayerTimeout'}}]


def test_remote_dumps_are_written():
    layout = """
        ##########
        #  b  y  #
        #a  ..  x#
        ##########
        """


    blue = FIXTURE_DIR / 'remote_dumps_are_written_blue.py'
    red = FIXTURE_DIR / 'remote_dumps_are_written_red.py'

    out_folder = tempfile.TemporaryDirectory()
    print(f"Using temporary folder to store the output: {out_folder}")

    state = pelita.game.run_game([str(blue), str(red)],
                                 max_rounds=2,
                                 layout_dict=pelita.layout.parse_layout(layout),
                                 store_output=out_folder.name)

    assert state['whowins'] == 2
    assert state['fatal_errors'] == [[], []]
    assert state['errors'] == [{}, {}]

    path = Path(out_folder.name)
    blue_lines = (path / 'blue.out').read_text().split('\n')
    red_lines = (path / 'red.out').read_text().split('\n')
    # now check what has been printed
    assert blue_lines == ['1 0 p1', '1 1 p1', '2 0 p1', '2 1 p1', '']
    assert red_lines == ['1 0 p2', '1 1 p2', '2 0 p2', '2 1 p2', '']

    assert (path / 'blue.err').read_text() == 'p1err\np1err\np1err\np1err\n'
    assert (path / 'red.err').read_text() == 'p2err\np2err\np2err\np2err\n'


@pytest.mark.parametrize("failing_team", [0, 1])
def test_remote_dumps_with_failure(failing_team):
    layout = """
        ##########
        #  b  y  #
        #a  ..  x#
        ##########
        """

    failing_player = FIXTURE_DIR / 'remote_dumps_with_failure_bad.py'
    good_player = FIXTURE_DIR / 'remote_dumps_with_failure_good.py'

    out_folder = tempfile.TemporaryDirectory()
    print(f"Using temporary folder to store the output: {out_folder}")

    if failing_team == 0:
        teams = [str(failing_player), str(good_player)]
    elif failing_team == 1:
        teams = [str(good_player), str(failing_player)]

    state = pelita.game.run_game(teams,
                                 max_rounds=2,
                                 layout_dict=pelita.layout.parse_layout(layout),
                                 store_output=out_folder.name)

    assert state['whowins'] == 1 - failing_team
    # when team 1 fails, it’s turn will be 1
    if failing_team == 0:
        fail_turn = 0
    elif failing_team == 1:
        fail_turn = 1
    assert state['fatal_errors'][failing_team][0] == {'type': 'FatalException',
                                           'description': 'Exception in client (ZeroDivisionError): division by zero',
                                           'turn': fail_turn,
                                           'round': 2}
    assert state['fatal_errors'][1 - failing_team] == []
    assert state['errors'] == [{}, {}]

    path = Path(out_folder.name)

    # stdout is still the same for both teams

    blue_lines = (path / 'blue.out').read_text().split('\n')
    red_lines = (path / 'red.out').read_text().split('\n')
    assert blue_lines == ['']
    assert red_lines == ['']

    blue_err = (path / 'blue.err').read_text()
    red_err = (path / 'red.err').read_text()

    if failing_team == 0:
        fail_err = blue_err
        good_err = red_err
    elif failing_team == 1:
        fail_err = red_err
        good_err = blue_err

    # Error is of the form:
    # Traceback (most recent call last):
    #   File ".../pelita/player/team.py", line 125, in get_move
    #     res = self._team_move(team[me.bot_turn], self._state)
    #   File "......", line 7, in move
    #     0 / 0
    # ZeroDivisionError: division by zero
    # -- empty line (index -1)

    assert "Traceback (most recent call last):" in fail_err
    assert "0 / 0" in fail_err
    assert "ZeroDivisionError: division by zero" in fail_err

    # No errors for red
    assert good_err == ""


@pytest.mark.parametrize("player_name,is_setup_error,error_type", [
    ['player_move_bad_type', False, 'ValueError'],
    ['player_move_bad_value', False, 'ValueError'],
    ['player_move_division_by_zero', False, 'ZeroDivisionError'],
    ['player_move_import_error', False, 'ModuleNotFoundError'],
    ['player_move_type_error', False, 'TypeError'],
    ['player_move_value_error', False, 'ValueError'],
    ['player_import_error', True, 'ModuleNotFoundError'],
    ['player_move_bad_args', False, 'TypeError'],
    ['player_move_bad_args_too_many', False, 'TypeError'],
    ['player_no_move', True, 'AttributeError'],
    ['player_no_name', True, 'AttributeError'],
    ['player_syntax_error', True, 'SyntaxError'],
])
def test_remote_move_failures(player_name, is_setup_error, error_type):
    layout = """
        ##########
        #  b  y  #
        #a  ..  x#
        ##########
        """

    failing_player = FIXTURE_DIR / player_name
    good_player = FIXTURE_DIR / 'remote_dumps_with_failure_good.py'

    if is_setup_error:
        state = pelita.game.run_game([str(failing_player), str(good_player)],
                                      max_rounds=2,
                                      layout_dict=pelita.layout.parse_layout(layout))

        assert state['whowins'] == 1

        assert state['fatal_errors'][0][0]['type'] == 'PlayerDisconnected'
        assert 'Could not load' in state['fatal_errors'][0][0]['description']
        assert error_type in state['fatal_errors'][0][0]['description']
        assert state['fatal_errors'][0][0]['turn'] == 0
        assert state['fatal_errors'][0][0]['round'] == None
        assert state['fatal_errors'][1] == []
        assert state['errors'] == [{}, {}]

    else:
        state = pelita.game.run_game([str(failing_player), str(good_player)],
                                      max_rounds=2,
                                      layout_dict=pelita.layout.parse_layout(layout))

        assert state['whowins'] == 1

        assert state['fatal_errors'][0][0]['type'] == 'FatalException'
        assert f'Exception in client ({error_type})' in state['fatal_errors'][0][0]['description']
        assert state['fatal_errors'][0][0]['turn'] == 0
        assert state['fatal_errors'][0][0]['round'] == 1
        assert state['fatal_errors'][1] == []
        assert state['errors'] == [{}, {}]
